"""Randomize the minitaur_gym_alternating_leg_env when reset() is called.

The randomization include swing_offset, extension_offset of all legs that mimics
bent legs, desired_pitch from user input, battery voltage and motor damping.
"""

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
parentdir = os.path.dirname(os.path.dirname(parentdir))
os.sys.path.insert(0, parentdir)

import numpy as np
import tf.compat.v1 as tf
from pybullet_envs.minitaur.envs import env_randomizer_base

# Absolute range.
NUM_LEGS = 4
BATTERY_VOLTAGE_RANGE = (14.8, 16.8)
MOTOR_VISCOUS_DAMPING_RANGE = (0, 0.01)


class MinitaurAlternatingLegsEnvRandomizer(env_randomizer_base.EnvRandomizerBase):
  """A randomizer that changes the minitaur_gym_alternating_leg_env."""

  def __init__(self,
               perturb_swing_bound=0.1,
               perturb_extension_bound=0.1,
               perturb_desired_pitch_bound=0.01):
    super(MinitaurAlternatingLegsEnvRandomizer, self).__init__()
    self.perturb_swing_bound = perturb_swing_bound
    self.perturb_extension_bound = perturb_extension_bound
    self.perturb_desired_pitch_bound = perturb_desired_pitch_bound

  def randomize_env(self, env):
    perturb_magnitude = np.random.uniform(low=-self.perturb_swing_bound,
                                          high=self.perturb_swing_bound,
                                          size=NUM_LEGS)
    env.set_swing_offset(perturb_magnitude)
    tf.logging.info("swing_offset: {}".format(perturb_magnitude))

    perturb_magnitude = np.random.uniform(low=-self.perturb_extension_bound,
                                          high=self.perturb_extension_bound,
                                          size=NUM_LEGS)
    env.set_extension_offset(perturb_magnitude)
    tf.logging.info("extension_offset: {}".format(perturb_magnitude))

    perturb_magnitude = np.random.uniform(low=-self.perturb_desired_pitch_bound,
                                          high=self.perturb_desired_pitch_bound)
    env.set_desired_pitch(perturb_magnitude)
    tf.logging.info("desired_pitch: {}".format(perturb_magnitude))

    randomized_battery_voltage = np.random.uniform(BATTERY_VOLTAGE_RANGE[0],
                                                   BATTERY_VOLTAGE_RANGE[1])
    env.minitaur.SetBatteryVoltage(randomized_battery_voltage)
    tf.logging.info("battery_voltage: {}".format(randomized_battery_voltage))

    randomized_motor_damping = np.random.uniform(MOTOR_VISCOUS_DAMPING_RANGE[0],
                                                 MOTOR_VISCOUS_DAMPING_RANGE[1])
    env.minitaur.SetMotorViscousDamping(randomized_motor_damping)
    tf.logging.info("motor_damping: {}".format(randomized_motor_damping))



"""This file implements the gym environment of example PyBullet simulation.

"""

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import time
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import pybullet
from pybullet_utils import bullet_client as bc

from pybullet_envs.prediction import boxstack_pybullet_sim

import os
import pybullet_data

from pkg_resources import parse_version


class PyBulletSimGymEnv(gym.Env):
  """The gym environment to run pybullet simulations.


  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 50}

  def __init__(self,
               pybullet_sim_factory=boxstack_pybullet_sim,
               render=True,
               render_sleep=False,
               debug_visualization=True,
               hard_reset=False,
               render_width=240,
               render_height=240,
               action_repeat=1,
               time_step=1. / 240.,
               num_bullet_solver_iterations=50,
               urdf_root=pybullet_data.getDataPath()):
    """Initialize the gym environment.

    Args:
      urdf_root: The path to the urdf data folder.
    """
    self._pybullet_sim_factory = pybullet_sim_factory
    self._time_step = time_step
    self._urdf_root = urdf_root
    self._observation = []
    self._action_repeat = action_repeat
    self._num_bullet_solver_iterations = num_bullet_solver_iterations
    self._env_step_counter = 0
    self._is_render = render
    self._debug_visualization = debug_visualization
    self._render_sleep = render_sleep
    self._render_width = render_width
    self._render_height = render_height
    self._cam_dist = .3
    self._cam_yaw = 50
    self._cam_pitch = -35
    self._hard_reset = True
    self._last_frame_time = 0.0

    optionstring = '--width={} --height={}'.format(render_width, render_height)

    print("urdf_root=" + self._urdf_root)

    if self._is_render:
      self._pybullet_client = bc.BulletClient(connection_mode=pybullet.GUI,
                                                         options=optionstring)
    else:
      self._pybullet_client = bc.BulletClient()

    if (debug_visualization == False):
      self._pybullet_client.configureDebugVisualizer(flag=self._pybullet_client.COV_ENABLE_GUI,
                                                     enable=0)
      self._pybullet_client.configureDebugVisualizer(
          flag=self._pybullet_client.COV_ENABLE_RGB_BUFFER_PREVIEW, enable=0)
      self._pybullet_client.configureDebugVisualizer(
          flag=self._pybullet_client.COV_ENABLE_DEPTH_BUFFER_PREVIEW, enable=0)
      self._pybullet_client.configureDebugVisualizer(
          flag=self._pybullet_client.COV_ENABLE_SEGMENTATION_MARK_PREVIEW, enable=0)

    self._pybullet_client.setAdditionalSearchPath(urdf_root)
    self.seed()
    self.reset()

    observation_high = (self._example_sim.GetObservationUpperBound())
    observation_low = (self._example_sim.GetObservationLowerBound())

    action_dim = self._example_sim.GetActionDimension()
    self._action_bound = 1
    action_high = np.array([self._action_bound] * action_dim)
    self.action_space = spaces.Box(-action_high, action_high)
    self.observation_space = spaces.Box(observation_low, observation_high)

    self.viewer = None
    self._hard_reset = hard_reset  # This assignment need to be after reset()

  def configure(self, args):
    self._args = args

  def reset(self):
    if self._hard_reset:
      self._pybullet_client.resetSimulation()

      self._pybullet_client.setPhysicsEngineParameter(
          numSolverIterations=int(self._num_bullet_solver_iterations))
      self._pybullet_client.setTimeStep(self._time_step)

      self._example_sim = self._pybullet_sim_factory.CreateSim(
          pybullet_client=self._pybullet_client,
          urdf_root=self._urdf_root,
          time_step=self._time_step)
    else:
      self._example_sim.Reset(reload_urdf=False)

    self._env_step_counter = 0

    #self._pybullet_client.resetDebugVisualizerCamera(
    #    self._cam_dist, self._cam_yaw, self._cam_pitch, [0, 0, 0])

    return self._get_observation()

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def step(self, action):
    """Step forward the simulation, given the action.

    Args:
      action: the predicted state

    Returns:
      observations: The actual state.
      reward: The reward for how well the prediction matches the actual state.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    if self._render_sleep:
      # Sleep, otherwise the computation takes less time than real time,
      # which will make the visualization like a fast-forward video.
      time_spent = time.time() - self._last_frame_time
      self._last_frame_time = time.time()
      time_to_sleep = self._action_repeat * self._time_step - time_spent
      if time_to_sleep > 0:
        time.sleep(time_to_sleep)

      #base_pos = self.minitaur.GetBasePosition()
      #self._pybullet_client.resetDebugVisualizerCamera(
      #    self._cam_dist, self._cam_yaw, self._cam_pitch, base_pos)

    for _ in range(self._action_repeat):
      self._example_sim.ApplyAction(action)
      self._pybullet_client.stepSimulation()

    self._env_step_counter += 1
    reward = self._reward()
    done = self._termination()
    return np.array(self._get_observation()), reward, done, {}

  def render(self, mode="rgb_array", close=False):
    if mode != "rgb_array":
      return np.array([])
    base_pos = [0, 0, 0]
    view_matrix = self._pybullet_client.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._cam_dist,
        yaw=self._cam_yaw,
        pitch=self._cam_pitch,
        roll=0,
        upAxisIndex=2)
    proj_matrix = self._pybullet_client.computeProjectionMatrixFOV(
        fov=60, aspect=float(self._render_width) / self._render_width, nearVal=0.01, farVal=100.0)
    proj_matrix = [
        1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1.0000200271606445, -1.0, 0.0, 0.0,
        -0.02000020071864128, 0.0
    ]
    (_, _, px, _, _) = self._pybullet_client.getCameraImage(
        width=self._render_width,
        height=self._render_height,
        viewMatrix=view_matrix,
        projectionMatrix=proj_matrix,
        renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)  #ER_TINY_RENDERER)
    rgb_array = np.array(px, dtype=np.uint8)
    rgb_array = np.reshape(rgb_array, (self._render_height, self._render_width, 4))
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def _termination(self):
    terminate = self._example_sim.Termination()
    return terminate

  def _reward(self):
    reward = 0
    return reward

  def _get_observation(self):
    self._observation = self._example_sim.GetObservation()
    return self._observation

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



"""Gait pattern planning module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import enum


class LegState(enum.Enum):
  """The state of a leg during locomotion."""
  SWING = 0
  STANCE = 1
  # A swing leg that collides with the ground.
  EARLY_CONTACT = 2
  # A stance leg that loses contact.
  LOSE_CONTACT = 3


class GaitGenerator(object):  # pytype: disable=ignored-metaclass
  """Generates the leg swing/stance pattern for the robot."""

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def reset(self, current_time):
    pass

  @abc.abstractmethod
  def update(self, current_time):
    pass



# Lint as: python3
"""The common class to manage the a stream of past data."""

import collections
from typing import Any, List, Sequence, Union
import dataclasses
import gin
import numpy as np


@dataclasses.dataclass
class BufferTuple:
  value_0: Any
  value_1: Any
  coeff: float


FloatOrArray = Union[float, Sequence[float]]
BufferTupleOrArray = Union[BufferTuple, Sequence[BufferTuple]]
TIME_IDX = 0
VALUE_IDX = 1


@gin.configurable
class TimeOrderedBuffer(object):
  """A buffer to hold and extract history data."""

  def __init__(self,
               max_buffer_timespan: float,
               error_on_timestamp_reversal: bool = True,
               error_on_duplicate_timestamp: bool = True,
               replace_value_on_duplicate_timestamp: bool = False,
               ):
    """Initializes the class.

    Args:
      max_buffer_timespan: Maximum amount of buffer by time to keep.
      error_on_timestamp_reversal: Whether to throw error if inverted timestamps
        are found.
      error_on_duplicate_timestamp: Whether to throw error if the incoming
        data has the same timestamp as the latest timestamp in the buffer.
      replace_value_on_duplicate_timestamp: Whether to keep the new value when
        a duplicate timestamp has occurred. This only applies if we are not
        throwing an error on duplicate timestamps.
    """
    if max_buffer_timespan < 0:
      raise ValueError(
          "Invalid max_buffer_timespan: {}".format(max_buffer_timespan))
    self._max_buffer_timespan = max_buffer_timespan
    self._error_on_timestamp_reversal = error_on_timestamp_reversal
    self._error_on_duplicate_timestamp = error_on_duplicate_timestamp
    self._replace_value_on_duplicate_timestamp = (
        replace_value_on_duplicate_timestamp)
    # TODO(tsangwei): Look for a ring buffer implementation.
    self._buffer = collections.deque()

  def reset(self):
    self._buffer.clear()

  def _compute_coeff(self,
                     newer_time: float,
                     older_time: float,
                     target_time: float,
                     ) -> float:
    """Compute the coefficient value between the two timestamps.

    Args:
      newer_time: The newer timestamp.
      older_time: The older timestamp.
      target_time: Target timestamp that is between the newer and older values.

    Returns:
      The coefficient as a float.
    """
    coeff = 0.0
    # Prevents divide by 0 error.
    if newer_time != older_time:
      assert older_time <= target_time <= newer_time
      coeff = (newer_time - target_time) / (newer_time - older_time)
    return coeff

  def _pack_data(self,
                 obs_newer: Any,
                 obs_older: Any,
                 target_time: float,
                 ) -> BufferTuple:
    """Packs up buffer data as BufferTuple dataclass.

    Args:
      obs_newer: Timestamp and value of newer observation.
      obs_older: Timestamp and value of older observation.
      target_time: Target timestamp of the observation we are looking for.

    Returns:
      BufferTuple dataclass.
    """
    coeff = self._compute_coeff(newer_time=obs_newer[TIME_IDX],
                                older_time=obs_older[TIME_IDX],
                                target_time=target_time)
    return BufferTuple(value_0=obs_newer[VALUE_IDX],
                       value_1=obs_older[VALUE_IDX],
                       coeff=coeff)

  def _find_values_at(self, timestamp_targets: Sequence[float]) -> List[Any]:
    """Get the lower/upper bound values for given target timestamp.

    Args:
      timestamp_targets: Actual timestamp value to match against.

    Returns:
      List of BufferTuple dataclass.
    """
    results = [None] * len(timestamp_targets)
    oldest_obs = self._buffer[0]
    latest_obs = self._buffer[-1]

    search_start_idx = None
    search_end_idx = None

    # Check to make sure we do not try to search for values outside of the
    # current buffer.
    for i in range(len(timestamp_targets)):
      # Oldest observation have the smallest timestamp.
      if timestamp_targets[i] <= oldest_obs[TIME_IDX]:
        results[i] = self._pack_data(obs_newer=oldest_obs,
                                     obs_older=oldest_obs,
                                     target_time=timestamp_targets[i])
      elif timestamp_targets[i] >= latest_obs[TIME_IDX]:
        results[i] = self._pack_data(obs_newer=latest_obs,
                                     obs_older=latest_obs,
                                     target_time=timestamp_targets[i])
      else:
        if search_end_idx is None:
          search_end_idx = i
        search_start_idx = i

    if search_end_idx is not None:
      results = self._walkthrough_buffer(timestamp_targets=timestamp_targets,
                                         search_start_idx=search_start_idx,
                                         search_end_idx=search_end_idx,
                                         results=results)

    return results

  def _walkthrough_buffer(self,
                          timestamp_targets: List[float],
                          search_start_idx: int,
                          search_end_idx: int,
                          results: List[BufferTuple],
                          ) -> List[BufferTuple]:
    """Actual method to walk through the buffer looking for requested values.

    Args:
      timestamp_targets: List of timestamps to search for in buffer.
      search_start_idx: Index number for timestamp_targets to start searching
        from.
      search_end_idx: Index number for timestamp_targets to stop searching at.
      results: List of BufferTuple values that covers out of bound results.

    Returns:
      List of BufferTuple values.
    """
    value_older = None
    target_idx = search_start_idx
    target_timestamp = timestamp_targets[target_idx]
    value_older = self._buffer[0]

    # Searching from oldest timestamp to latest timestamp.
    for value_newer in self._buffer:
      # Catch edge case scenario where multiple timestamp targets are between
      # the same two buffer timestamps. (b/157104935)
      while value_newer[TIME_IDX] >= target_timestamp:
        # Catch special edge case scenario if using older_obs_blender method.
        obs_older = value_newer if (
            value_newer[TIME_IDX] == target_timestamp) else value_older
        results[target_idx] = self._pack_data(obs_newer=value_newer,
                                              obs_older=obs_older,
                                              target_time=target_timestamp)
        if target_idx - 1 >= search_end_idx:
          target_idx -= 1
          target_timestamp = timestamp_targets[target_idx]
        else:
          return results
      value_older = value_newer

    return results

  def add(self, timestamp: float, value: Any):
    """Inserts timestamp and value into buffer.

    Args:
      timestamp: Timestamp of the data value.
      value: Data value to be saved into the buffer.
    """
    if self._buffer:
      last_timestamp = self._buffer[-1][TIME_IDX]
      if last_timestamp == timestamp:
        if (not np.array_equal(self._buffer[-1][VALUE_IDX], value) and
            self._error_on_duplicate_timestamp):
          raise ValueError("Duplicate timestamp detected: {}".format(timestamp))
        else:
          # Duplicate message detected.
          if self._replace_value_on_duplicate_timestamp:
            self._buffer[-1] = (timestamp, value)
          return
      if last_timestamp > timestamp and self._error_on_timestamp_reversal:
        raise ValueError(
            "Time reversal detected: new timestamp is {} vs last timestamp {}"
            .format(timestamp, last_timestamp))
      # Dropping old buffer data that exceed buffer timespan limit and making
      # sure the buffer does not go empty.
      while (len(self._buffer) > 1 and self._max_buffer_timespan <
             (timestamp - self._buffer[1][TIME_IDX])):
        self._buffer.popleft()
    self._buffer.append((timestamp, value))

  def get_delayed_value(self, latency: FloatOrArray) -> BufferTupleOrArray:
    """Retrieves value in the history buffer according to latency.

    Finds the closest pair of values that are some time (i.e. latency) ago from
    the most recent timestamp. Suppose the history buffer looks like this:

    [(0, val_x),...,(0.6, val_k-1), (0.7, val_k), (0.8, val_k+1),...,(2, val_0)]

    And the latency is '1.33', then this API will locate the values with
    timestamps close to 2 - 1.33 = 0.67. So in this case, it will return the
    pair (0.7, val_k) and (0.6, val_k+1), as well as a blending coefficient
    which is calculated by (0.7 - 0.67) / (0.7 - 0.6) = 0.3. This blending coeff
    can be used to linearly interpolate the returned values, i.e. val_1 * (1 -
    coeff) + val_2 * coeff, if the multiply operator is defined.

    Args:
      latency: The time interval(s) to look backwards in the history buffer from
        the most recent timestamp.

    Returns:
      An array of BufferTuple dataclass.

    Raises:
      ValueError: if the latency is negative.
      BufferError: if the buffer is empty.
    """
    buffer_len = len(self._buffer)
    if buffer_len == 0:
      raise BufferError("The buffer is empty. Have you called 'add'?")

    single_latency = isinstance(latency, (int, float))

    if single_latency:
      if latency < 0:
        raise ValueError("Latency cannot be negative.")
    else:
      if any(value < 0 for value in latency):
        raise ValueError("Latency list contains negative values.")
      if latency != sorted(latency):
        raise ValueError("Invalid unsorted latency list.")

    target_list = [latency] if single_latency else latency
    current_time = self._buffer[-1][TIME_IDX]
    target_list = [current_time - offset for offset in target_list]

    result = self._find_values_at(target_list)
    return result[0] if single_latency else result



"""Denormalize the action from [-1, 1] to the env.action_space."""

import gin
import gym
import numpy as np


def _denomralize(env, action):
  action = np.array(action)
  low = np.array(env.action_space.low)
  high = np.array(env.action_space.high)
  return (high - low) / 2.0 * action + (high + low) / 2.0


@gin.configurable
class ActionDenormalizeWrapper(object):
  """An env wrapper that denormalize the action from [-1, 1] to the bounds."""

  def __init__(self, gym_env):
    """Initializes the wrapper."""
    self._gym_env = gym_env
    self.action_space = gym.spaces.Box(
        low=-1.0,
        high=1.0,
        shape=self._gym_env.action_space.low.shape,
        dtype=np.float32)

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array between [-1.0, 1.0].

    Returns:
      The tuple containing the observation, the reward, and the epsiode
        end indicator.
    """
    return self._gym_env.step(_denomralize(self._gym_env, action))



import numpy as np


class PDControllerStableMultiDof(object):

  def __init__(self, pb):
    self._pb = pb

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):

    numJoints = len(jointIndices)  #self._pb.getNumJoints(bodyUniqueId)
    curPos, curOrn = self._pb.getBasePositionAndOrientation(bodyUniqueId)
    #q1 = [desiredPositions[0],desiredPositions[1],desiredPositions[2],desiredPositions[3],desiredPositions[4],desiredPositions[5],desiredPositions[6]]
    q1 = [curPos[0], curPos[1], curPos[2], curOrn[0], curOrn[1], curOrn[2], curOrn[3]]

    #qdot1 = [0,0,0, 0,0,0,0]
    baseLinVel, baseAngVel = self._pb.getBaseVelocity(bodyUniqueId)

    qdot1 = [
        baseLinVel[0], baseLinVel[1], baseLinVel[2], baseAngVel[0], baseAngVel[1], baseAngVel[2], 0
    ]
    qError = [0, 0, 0, 0, 0, 0, 0]

    qIndex = 7
    qdotIndex = 7
    zeroAccelerations = [0, 0, 0, 0, 0, 0, 0]
    for i in range(numJoints):
      js = self._pb.getJointStateMultiDof(bodyUniqueId, jointIndices[i])

      jointPos = js[0]
      jointVel = js[1]
      q1 += jointPos

      if len(js[0]) == 1:
        desiredPos = desiredPositions[qIndex]

        qdiff = desiredPos - jointPos[0]
        qError.append(qdiff)
        zeroAccelerations.append(0.)
        qdot1 += jointVel
        qIndex += 1
        qdotIndex += 1
      if len(js[0]) == 4:
        desiredPos = [
            desiredPositions[qIndex], desiredPositions[qIndex + 1], desiredPositions[qIndex + 2],
            desiredPositions[qIndex + 3]
        ]
        axis = self._pb.getAxisDifferenceQuaternion(desiredPos, jointPos)
        jointVelNew = [jointVel[0], jointVel[1], jointVel[2], 0]
        qdot1 += jointVelNew
        qError.append(axis[0])
        qError.append(axis[1])
        qError.append(axis[2])
        qError.append(0)
        desiredVel = [
            desiredVelocities[qdotIndex], desiredVelocities[qdotIndex + 1],
            desiredVelocities[qdotIndex + 2]
        ]
        zeroAccelerations += [0., 0., 0., 0.]
        qIndex += 4
        qdotIndex += 4

    q = np.array(q1)
    qdot = np.array(qdot1)

    qdotdesired = np.array(desiredVelocities)
    qdoterr = qdotdesired - qdot

    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)

    # Compute -Kp(q + qdot - qdes)
    p_term = Kp.dot(qError - qdot*timeStep)
    # Compute -Kd(qdot - qdotdes)
    d_term = Kd.dot(qdoterr)

    # Compute Inertia matrix M(q)
    M = self._pb.calculateMassMatrix(bodyUniqueId, q1, flags=1)
    M = np.array(M)
    # Given: M(q) * qddot + C(q, qdot) = T_ext + T_int
    # Compute Coriolis and External (Gravitational) terms G = C - T_ext
    G = self._pb.calculateInverseDynamics(bodyUniqueId, q1, qdot1, zeroAccelerations, flags=1)
    G = np.array(G)
    # Obtain estimated generalized accelerations, considering Coriolis and Gravitational forces, and stable PD actions
    qddot = np.linalg.solve(a=(M + Kd * timeStep),
                            b=p_term + d_term - G)
    # Compute control generalized forces (T_int)
    tau = p_term + d_term - Kd.dot(qddot) * timeStep
    # Clip generalized forces to actuator limits
    maxF = np.array(maxForces)
    generalized_forces = np.clip(tau, -maxF, maxF)
    return generalized_forces


class PDControllerStable(object):
  """
  Implementation based on: Tan, J., Liu, K., & Turk, G. (2011). "Stable proportional-derivative controllers"
  DOI: 10.1109/MCG.2011.30
  """
  def __init__(self, pb):
    self._pb = pb

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):
    numJoints = self._pb.getNumJoints(bodyUniqueId)
    jointStates = self._pb.getJointStates(bodyUniqueId, jointIndices)
    q1 = []
    qdot1 = []
    zeroAccelerations = []
    for i in range(numJoints):
      q1.append(jointStates[i][0])
      qdot1.append(jointStates[i][1])
      zeroAccelerations.append(0)

    q = np.array(q1)
    qdot = np.array(qdot1)
    qdes = np.array(desiredPositions)
    qdotdes = np.array(desiredVelocities)

    qError = qdes - q
    qdotError = qdotdes - qdot

    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)

    # Compute -Kp(q + qdot - qdes)
    p_term = Kp.dot(qError - qdot*timeStep)
    # Compute -Kd(qdot - qdotdes)
    d_term = Kd.dot(qdotError)

    # Compute Inertia matrix M(q)
    M = self._pb.calculateMassMatrix(bodyUniqueId, q1)
    M = np.array(M)
    # Given: M(q) * qddot + C(q, qdot) = T_ext + T_int
    # Compute Coriolis and External (Gravitational) terms G = C - T_ext
    G = self._pb.calculateInverseDynamics(bodyUniqueId, q1, qdot1, zeroAccelerations)
    G = np.array(G)
    # Obtain estimated generalized accelerations, considering Coriolis and Gravitational forces, and stable PD actions
    qddot = np.linalg.solve(a=(M + Kd * timeStep),
                            b=(-G + p_term + d_term))
    # Compute control generalized forces (T_int)
    tau = p_term + d_term - (Kd.dot(qddot) * timeStep)
    # Clip generalized forces to actuator limits
    maxF = np.array(maxForces)
    generalized_forces = np.clip(tau, -maxF, maxF)
    return generalized_forces



import numpy as np
from pybullet_envs.deep_mimic.env.env import Env


class Path(object):

  def __init__(self):
    self.clear()
    return

  def pathlength(self):
    return len(self.actions)

  def is_valid(self):
    valid = True
    l = self.pathlength()
    valid &= len(self.states) == l + 1
    valid &= len(self.goals) == l + 1
    valid &= len(self.actions) == l
    valid &= len(self.logps) == l
    valid &= len(self.rewards) == l
    valid &= len(self.flags) == l

    return valid

  def check_vals(self):
    for vals in [self.states, self.goals, self.actions, self.logps, self.rewards]:
      for v in vals:
        if not np.isfinite(v).all():
          return False
    return True

  def clear(self):
    self.states = []
    self.goals = []
    self.actions = []
    self.logps = []
    self.rewards = []
    self.flags = []
    self.terminate = Env.Terminate.Null
    return

  def get_pathlen(self):
    return len(self.rewards)

  def calc_return(self):
    return sum(self.rewards)



import re as RE


class ArgParser(object):
  global_parser = None

  def __init__(self):
    self._table = dict()
    return

  def clear(self):
    self._table.clear()
    return

  def load_args(self, arg_strs):
    succ = True
    vals = []
    curr_key = ''

    for str in arg_strs:
      if not (self._is_comment(str)):
        is_key = self._is_key(str)
        if (is_key):
          if (curr_key != ''):
            if (curr_key not in self._table):
              self._table[curr_key] = vals

          vals = []
          curr_key = str[2::]
        else:
          vals.append(str)

    if (curr_key != ''):
      if (curr_key not in self._table):
        self._table[curr_key] = vals

      vals = []

    return succ

  def load_file(self, filename):
    succ = False
    with open(filename, 'r') as file:
      lines = RE.split(r'[\n\r]+', file.read())
      file.close()

      arg_strs = []
      for line in lines:
        if (len(line) > 0 and not self._is_comment(line)):
          arg_strs += line.split()

      succ = self.load_args(arg_strs)
    return succ

  def has_key(self, key):
    return key in self._table

  def parse_string(self, key, default=''):
    str = default
    if self.has_key(key):
      str = self._table[key][0]
    return str

  def parse_strings(self, key, default=[]):
    arr = default
    if self.has_key(key):
      arr = self._table[key]
    return arr

  def parse_int(self, key, default=0):
    val = default
    if self.has_key(key):
      val = int(self._table[key][0])
    return val

  def parse_ints(self, key, default=[]):
    arr = default
    if self.has_key(key):
      arr = [int(str) for str in self._table[key]]
    return arr

  def parse_float(self, key, default=0.0):
    val = default
    if self.has_key(key):
      val = float(self._table[key][0])
    return val

  def parse_floats(self, key, default=[]):
    arr = default
    if self.has_key(key):
      arr = [float(str) for str in self._table[key]]
    return arr

  def parse_bool(self, key, default=False):
    val = default
    if self.has_key(key):
      val = self._parse_bool(self._table[key][0])
    return val

  def parse_bools(self, key, default=[]):
    arr = default
    if self.has_key(key):
      arr = [self._parse_bool(str) for str in self._table[key]]
    return arr

  def _is_comment(self, str):
    is_comment = False
    if (len(str) > 0):
      is_comment = str[0] == '#'

    return is_comment

  def _is_key(self, str):
    is_key = False
    if (len(str) >= 3):
      is_key = str[0] == '-' and str[1] == '-'

    return is_key

  def _parse_bool(self, str):
    val = False
    if (str == 'true' or str == 'True' or str == '1' or str == 'T' or str == 't'):
      val = True
    return val



# Lint as: python3
"""Converts a list of sensors to gym space."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import List
import gin
import gym
from gym import spaces
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor


class UnsupportedConversionError(NotImplementedError):
  """An exception when the function cannot convert sensors to the gym space."""


class AmbiguousDataTypeError(TypeError):
  """An exception when the function cannot determine the data type."""


@gin.configurable
def convert_sensors_to_gym_space(sensors: List[sensor.Sensor]) -> gym.Space:
  """Convert a list of sensors to the corresponding gym space.

  Args:
    sensors: a list of the current sensors

  Returns:
    space: the converted gym space

  Raises:
    UnsupportedConversionError: raises when the function cannot convert the
      given list of sensors.
  """

  if all([
      isinstance(s, sensor.BoxSpaceSensor) and s.get_dimension() == 1
      for s in sensors
  ]):
    return convert_1d_box_sensors_to_gym_space(sensors)
  raise UnsupportedConversionError('sensors = ' + str(sensors))


@gin.configurable
def convert_1d_box_sensors_to_gym_space(
    sensors: List[sensor.Sensor]) -> gym.Space:
  """Convert a list of 1D BoxSpaceSensors to the corresponding gym space.

  Args:
    sensors: a list of the current sensors

  Returns:
    space: the converted gym space

  Raises:
    UnsupportedConversionError: raises when the function cannot convert the
      given list of sensors.
    AmbiguousDataTypeError: raises when the function cannot determine the
      data types because they are not uniform.
  """
  # Check if all sensors are 1D BoxSpaceSensors
  if not all([
      isinstance(s, sensor.BoxSpaceSensor) and s.get_dimension() == 1
      for s in sensors
  ]):
    raise UnsupportedConversionError('sensors = ' + str(sensors))

  # Check if all sensors have the same data type
  sensor_dtypes = [s.get_dtype() for s in sensors]
  if sensor_dtypes.count(sensor_dtypes[0]) != len(sensor_dtypes):
    raise AmbiguousDataTypeError('sensor datatypes are inhomogeneous')

  lower_bound = np.concatenate([s.get_lower_bound() for s in sensors])
  upper_bound = np.concatenate([s.get_upper_bound() for s in sensors])
  observation_space = spaces.Box(
      np.array(lower_bound), np.array(upper_bound), dtype=np.float32)
  return observation_space


@gin.configurable
def convert_sensors_to_gym_space_dictionary(
    sensors: List[sensor.Sensor]) -> gym.Space:
  """Convert a list of sensors to the corresponding gym space dictionary.

  Args:
    sensors: a list of the current sensors

  Returns:
    space: the converted gym space dictionary

  Raises:
    UnsupportedConversionError: raises when the function cannot convert the
      given list of sensors.
  """
  gym_space_dict = {}
  for s in sensors:
    if isinstance(s, sensor.BoxSpaceSensor):
      gym_space_dict[s.get_name()] = spaces.Box(
          np.array(s.get_lower_bound()),
          np.array(s.get_upper_bound()),
          dtype=np.float32)
    elif isinstance(s, sensor.Sensor):
      if isinstance(s.observation_space, spaces.Box):
        gym_space_dict[s.get_name()] = s.observation_space
      elif isinstance(s.observation_space, spaces.Dict):
        gym_space_dict.update(s.observation_space.spaces)
      else:
        raise UnsupportedConversionError(
            f'Unsupported space type {type(s.observation_space)}, '
            f'must be Box or Dict. sensor = {s}')
    else:
      raise UnsupportedConversionError('sensors = ' + str(sensors))
  return spaces.Dict(gym_space_dict)


def create_constant_action(action_space, action_value=0):
  """Create an uniform value action based on the type of action space."""
  if isinstance(action_space, gym.spaces.Dict):
    # gym.spaces.Dict has a shape of None, so construct action over subspaces.
    return {
        sub_name: create_constant_action(sub_space, action_value)
        for sub_name, sub_space in action_space.spaces.items()
    }
  else:  # Presumably gym.spaces.Box, but in case it is not ...
    return np.full(action_space.shape, action_value)


def action_astype(action, dtype):
  """Transform an action to a different datatype."""
  if isinstance(action, dict):
    return {key: action_astype(value, dtype) for key, value in action.items()}
  else:
    return np.array(action, dtype=dtype)



import pybullet as p
from time import sleep
import pybullet_data

physicsClient = p.connect(p.GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())

p.resetSimulation(p.RESET_USE_DEFORMABLE_WORLD)
p.resetDebugVisualizerCamera(3,-420,-30,[0.3,0.9,-2])
p.setGravity(0, 0, -10)

tex = p.loadTexture("uvmap.png")
planeId = p.loadURDF("plane.urdf", [0,0,-2])

boxId = p.loadURDF("cube.urdf", [0,3,2],useMaximalCoordinates = True)

bunnyId = p.loadSoftBody("torus/torus_textured.obj", simFileName="torus.vtk", mass = 3, useNeoHookean = 1, NeoHookeanMu = 180, NeoHookeanLambda = 600, NeoHookeanDamping = 0.01, collisionMargin = 0.006, useSelfCollision = 1, frictionCoeff = 0.5, repulsionStiffness = 800)
p.changeVisualShape(bunnyId, -1, rgbaColor=[1,1,1,1], textureUniqueId=tex, flags=0)


bunny2 = p.loadURDF("torus_deform.urdf", [0,1,0.2], flags=p.URDF_USE_SELF_COLLISION)

p.changeVisualShape(bunny2, -1, rgbaColor=[1,1,1,1], textureUniqueId=tex, flags=0)
p.setPhysicsEngineParameter(sparseSdfVoxelSize=0.25)
p.setRealTimeSimulation(0)

while p.isConnected():
  p.stepSimulation()
  p.getCameraImage(320,200)
  p.setGravity(0,0,-10)



"""see https://machinelearningmastery.com/multivariate-time-series-forecasting-lstms-keras/"""

import matplotlib
matplotlib.use('TkAgg')
import numpy as np
from matplotlib import pyplot
from pandas import read_csv
from pandas import DataFrame
from pandas import concat
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import mean_squared_error
from keras.models import Sequential
from keras.layers import Dropout
from keras.layers import Dense
from keras.layers import LSTM
from keras.callbacks import ModelCheckpoint
from matplotlib import pyplot

from pandas import read_csv


# convert series to supervised learning
def series_to_supervised(data, n_in=1, n_out=1, dropnan=True):
  n_vars = 1 if type(data) is list else data.shape[1]
  df = DataFrame(data)
  cols, names = list(), list()
  # input sequence (t-n, ... t-1)
  for i in range(n_in, 0, -1):
    cols.append(df.shift(i))
    names += [('var%d(t-%d)' % (j + 1, i)) for j in range(n_vars)]
  # forecast sequence (t, t+1, ... t+n)
  for i in range(0, n_out):
    cols.append(df.shift(-i))
    if i == 0:
      names += [('var%d(t)' % (j + 1)) for j in range(n_vars)]
    else:
      names += [('var%d(t+%d)' % (j + 1, i)) for j in range(n_vars)]
  # put it all together
  agg = concat(cols, axis=1)
  agg.columns = names
  # drop rows with NaN values
  if dropnan:
    agg.dropna(inplace=True)
  return agg


values = [x for x in range(10)]
data = series_to_supervised(values, 2)
print(data)

# load dataset
#dataset = read_csv('./data/minitaur_log_latency_0.01.csv')
#dataset = read_csv('./data/minitaur_log_latency_0.003.csv')
dataset = read_csv('./data/minitaur_log_latency_0.006.csv')

values = dataset.values
# integer encode direction
#encoder = LabelEncoder()
#values[:,3] = encoder.fit_transform(values[:,3])
# ensure all data is float
values = values.astype('float32')

# normalize features
useNormalization = False
if useNormalization:
  scaler = MinMaxScaler(feature_range=(0, 1))
  scaled = scaler.fit_transform(values)
else:
  scaled = values

# frame as supervised learning
lag_steps = 5
reframed = series_to_supervised(scaled, lag_steps, 1)
print("reframed before drop=", reframed)

# drop columns we don't want to predict
reframed.drop(reframed.columns[[3, 7, 11, 15, 19]], axis=1, inplace=True)
print("after drop=", reframed.head())

#dummy = scaler.inverse_transform(reframed)
#print(dummy)

groups = [0, 1, 2, 3]

i = 1
# plot each column
doPlot = False
if doPlot:
  pyplot.figure()
  for group in groups:
    pyplot.subplot(len(groups), 1, i)
    pyplot.plot(values[0:25, group])
    pyplot.title(dataset.columns[group], y=0.5, loc='right')
    i += 1
  pyplot.show()

# split into train and test sets
values = reframed.values
n_train_hours = 6000
train = values[:n_train_hours, :]
test = values[n_train_hours:, :]
# split into input and outputs
train_X, train_y = train[:, :-1], train[:, -1]
test_X, test_y = test[:, :-1], test[:, -1]

print("train_X.shape[1]=", train_X.shape[1])

# design network
useLSTM = True
if useLSTM:
  # reshape input to be 3D [samples, timesteps, features]
  train_X = train_X.reshape(
      (train_X.shape[0], lag_steps + 1, int(train_X.shape[1] / (lag_steps + 1))))
  test_X = test_X.reshape((test_X.shape[0], lag_steps + 1, int(test_X.shape[1] / (lag_steps + 1))))
  model = Sequential()
  model.add(LSTM(40, input_shape=(train_X.shape[1], train_X.shape[2])))
  model.add(Dropout(0.05))
  model.add(Dense(8, activation='sigmoid'))
  model.add(Dense(8, activation='sigmoid'))
  model.add(Dropout(0.05))
  model.add(Dense(1, activation='linear'))
else:
  # create model
  model = Sequential()
  model.add(Dense(12, input_dim=train_X.shape[1], activation="relu"))
  model.add(Dense(8, activation="sigmoid"))
  model.add(Dense(1, activation="linear"))

#model.compile(loss='mae', optimizer='adam')
model.compile(loss='mean_squared_error', optimizer='adam', metrics=['mean_squared_error'])

# checkpoint
filepath = '/tmp/keras/weights-improvement-{epoch:02d}-{val_loss:.2f}.hdf5'
checkpoint = ModelCheckpoint(filepath,
                             monitor='val_loss',
                             verbose=1,
                             save_best_only=True,
                             mode='min')
callbacks_list = [checkpoint]

# fit network
history = model.fit(train_X,
                    train_y,
                    epochs=1500,
                    batch_size=32,
                    callbacks=callbacks_list,
                    validation_data=(test_X, test_y),
                    verbose=2,
                    shuffle=False)
# plot history

data = np.array([[[
    1.513535008329887299, 3.234624992847829894e-01, 1.731481043119239782, 1.741165415165205399,
    1.534267104753672228e+00, 1.071354965017878635e+00, 1.712386127673626302e+00
]]])

#prediction = model.predict(data)
#print("prediction=",prediction)

pyplot.plot(history.history['loss'], label='train')
pyplot.plot(history.history['val_loss'], label='test')
pyplot.legend()
pyplot.show()



import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import time
import pybullet
from pybullet_utils import bullet_client as bc
from . import racecar
import random
import pybullet_data
from pkg_resources import parse_version

RENDER_HEIGHT = 720
RENDER_WIDTH = 960


class RacecarZEDGymEnv(gym.Env):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self,
               urdfRoot=pybullet_data.getDataPath(),
               actionRepeat=10,
               isEnableSelfCollision=True,
               isDiscrete=False,
               renders=True):
    print("init")
    self._timeStep = 0.01
    self._urdfRoot = urdfRoot
    self._actionRepeat = actionRepeat
    self._isEnableSelfCollision = isEnableSelfCollision
    self._ballUniqueId = -1
    self._envStepCounter = 0
    self._renders = renders
    self._width = 100
    self._height = 10

    self._isDiscrete = isDiscrete
    if self._renders:
      self._p = bc.BulletClient(connection_mode=pybullet.GUI)
    else:
      self._p = bc.BulletClient()

    self.seed()
    self.reset()
    observationDim = len(self.getExtendedObservation())
    #print("observationDim")
    #print(observationDim)

    observation_high = np.array([np.finfo(np.float32).max] * observationDim)
    if (isDiscrete):
      self.action_space = spaces.Discrete(9)
    else:
      action_dim = 2
      self._action_bound = 1
      action_high = np.array([self._action_bound] * action_dim)
      self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self.observation_space = spaces.Box(low=0,
                                        high=255,
                                        shape=(self._height, self._width, 4),
                                        dtype=np.uint8)

    self.viewer = None

  def reset(self):
    self._p.resetSimulation()
    #p.setPhysicsEngineParameter(numSolverIterations=300)
    self._p.setTimeStep(self._timeStep)
    #self._p.loadURDF(os.path.join(os.path.dirname(__file__),"../data","plane.urdf"))
    stadiumobjects = self._p.loadSDF(os.path.join(self._urdfRoot, "stadium.sdf"))
    #move the stadium objects slightly above 0
    for i in stadiumobjects:
      pos, orn = self._p.getBasePositionAndOrientation(i)
      newpos = [pos[0], pos[1], pos[2] + 0.1]
      self._p.resetBasePositionAndOrientation(i, newpos, orn)

    dist = 5 + 2. * random.random()
    ang = 2. * 3.1415925438 * random.random()

    ballx = dist * math.sin(ang)
    bally = dist * math.cos(ang)
    ballz = 1

    self._ballUniqueId = self._p.loadURDF(os.path.join(self._urdfRoot, "sphere2red.urdf"),
                                          [ballx, bally, ballz])
    self._p.setGravity(0, 0, -10)
    self._racecar = racecar.Racecar(self._p, urdfRootPath=self._urdfRoot, timeStep=self._timeStep)
    self._envStepCounter = 0
    for i in range(100):
      self._p.stepSimulation()
    self._observation = self.getExtendedObservation()
    return np.array(self._observation)

  def __del__(self):
    self._p = 0

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def getExtendedObservation(self):
    carpos, carorn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
    carmat = self._p.getMatrixFromQuaternion(carorn)
    ballpos, ballorn = self._p.getBasePositionAndOrientation(self._ballUniqueId)
    invCarPos, invCarOrn = self._p.invertTransform(carpos, carorn)
    ballPosInCar, ballOrnInCar = self._p.multiplyTransforms(invCarPos, invCarOrn, ballpos, ballorn)
    dist0 = 0.3
    dist1 = 1.
    eyePos = [
        carpos[0] + dist0 * carmat[0], carpos[1] + dist0 * carmat[3],
        carpos[2] + dist0 * carmat[6] + 0.3
    ]
    targetPos = [
        carpos[0] + dist1 * carmat[0], carpos[1] + dist1 * carmat[3],
        carpos[2] + dist1 * carmat[6] + 0.3
    ]
    up = [carmat[2], carmat[5], carmat[8]]
    viewMat = self._p.computeViewMatrix(eyePos, targetPos, up)
    #viewMat = self._p.computeViewMatrixFromYawPitchRoll(carpos,1,0,0,0,2)
    #print("projectionMatrix:")
    #print(self._p.getDebugVisualizerCamera()[3])
    projMatrix = [
        0.7499999403953552, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, -1.0000200271606445, -1.0,
        0.0, 0.0, -0.02000020071864128, 0.0
    ]
    img_arr = self._p.getCameraImage(width=self._width,
                                     height=self._height,
                                     viewMatrix=viewMat,
                                     projectionMatrix=projMatrix)
    rgb = img_arr[2]
    np_img_arr = np.reshape(rgb, (self._height, self._width, 4))
    self._observation = np_img_arr
    return self._observation

  def step(self, action):
    if (self._renders):
      basePos, orn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
      #self._p.resetDebugVisualizerCamera(1, 30, -40, basePos)

    if (self._isDiscrete):
      fwd = [-1, -1, -1, 0, 0, 0, 1, 1, 1]
      steerings = [-0.6, 0, 0.6, -0.6, 0, 0.6, -0.6, 0, 0.6]
      forward = fwd[action]
      steer = steerings[action]
      realaction = [forward, steer]
    else:
      realaction = action

    self._racecar.applyAction(realaction)
    for i in range(self._actionRepeat):
      self._p.stepSimulation()
      if self._renders:
        time.sleep(self._timeStep)
      self._observation = self.getExtendedObservation()

      if self._termination():
        break
      self._envStepCounter += 1
    reward = self._reward()
    done = self._termination()
    #print("len=%r" % len(self._observation))

    return np.array(self._observation), reward, done, {}

  def render(self, mode='human', close=False):
    if mode != "rgb_array":
      return np.array([])
    base_pos, orn = self._p.getBasePositionAndOrientation(self._racecar.racecarUniqueId)
    view_matrix = self._p.computeViewMatrixFromYawPitchRoll(cameraTargetPosition=base_pos,
                                                            distance=self._cam_dist,
                                                            yaw=self._cam_yaw,
                                                            pitch=self._cam_pitch,
                                                            roll=0,
                                                            upAxisIndex=2)
    proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
                                                     aspect=float(RENDER_WIDTH) / RENDER_HEIGHT,
                                                     nearVal=0.1,
                                                     farVal=100.0)
    (_, _, px, _, _) = self._p.getCameraImage(width=RENDER_WIDTH,
                                              height=RENDER_HEIGHT,
                                              viewMatrix=view_matrix,
                                              projectionMatrix=proj_matrix,
                                              renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)
    rgb_array = np.array(px)
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def _termination(self):
    return self._envStepCounter > 1000

  def _reward(self):
    closestPoints = self._p.getClosestPoints(self._racecar.racecarUniqueId, self._ballUniqueId,
                                             10000)

    numPt = len(closestPoints)
    reward = -1000
    #print(numPt)
    if (numPt > 0):
      #print("reward:")
      reward = -closestPoints[0][8]
      #print(reward)
    return reward

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



r"""An example to run an OpenAI gym environment with laikago and PMTG wrapper.

This is an open-loop controller where we use 0 residuals and only execute the
trajectory generator. The parameters that are (normally modulated by the policy)
are fixed (except the intensity) and not optimized. They are hand picked as
follows:
 - The gait cycle frequency is 3 Hz.
 - Walking height is neutral (0).
 - Swing vs stance ratio is 2 (swing takes half the time vs stance).
 - Intensity starts from zero and is gradually increased over time.

blaze run -c opt \
//robotics/reinforcement_learning/minitaur/envs_v2/examples\
:laikago_pmtg_example
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import gin
import sys

from pybullet_envs.minitaur.envs_v2 import env_loader
import pybullet_data as pd

CONFIG_DIR = pd.getDataPath()+"/configs_v2/"
CONFIG_FILES = [
    os.path.join(CONFIG_DIR, "base/laikago_with_imu.gin"),
    os.path.join(CONFIG_DIR, "tasks/fwd_task_no_termination.gin"),
    os.path.join(CONFIG_DIR, "wrappers/pmtg_wrapper.gin"),
    os.path.join(CONFIG_DIR, "scenes/simple_scene.gin")
]
# Constants used to show example PMTG behavior with zero residual.
# Since we use the default PMTG configuration, there is only one trajectory
# generator. So the parameters that can be changed are:
# - Multiplier for delta time (similar to frequency but per time step).
# - Intensity of the trajectory generator.
# - Walking heights used for the legs.
# - The ratio of the speed of the leg during swing vs stance phase.
# For more details, check out TgSimple._process_tg_params method.
_PMTG_DELTA_TIME_MULTIPLIER = 2.0
_PMTG_INTENSITY_RANGE = (0.0, 1.5)
_PMTG_INTENSITY_STEP_SIZE = 0.0001
_PMTG_WALKING_HEIGHT = -0.3
_PMTG_SWING_VS_STANCE = 2
_NUM_MOTORS = 12


def main(argv):
  del argv  # Unused.

  # Parse the gym config and create the environment.
  for gin_file in CONFIG_FILES:
    gin.parse_config_file(gin_file)
  gin.bind_parameter("SimulationParameters.enable_rendering", True)
  gin.bind_parameter("terminal_conditions.maxstep_terminal_condition.max_step",
                     10000)
  env = env_loader.load()
  tg_intensity = _PMTG_INTENSITY_RANGE[0]
  sum_reward = 0
  env.reset()
  done = False
  # Use zero residual, only use the output of the trajectory generator.
  residual = [0] * _NUM_MOTORS
  # Since we fix residuals and all the parameters of the TG, this example
  # is practically an open loop controller. A learned policy would provide
  # different values for these parameters at every timestep.
  while not done:
    # Increase the intensity of the trajectory generator gradually
    # to illustrate increasingly larger steps.
    if tg_intensity < _PMTG_INTENSITY_RANGE[1]:
      tg_intensity += _PMTG_INTENSITY_STEP_SIZE
    tg_params = [
        _PMTG_DELTA_TIME_MULTIPLIER, tg_intensity, _PMTG_WALKING_HEIGHT,
        _PMTG_SWING_VS_STANCE
    ]
    action = residual + tg_params
    _, reward, done, _ = env.step(action)
    sum_reward += reward


if __name__ == "__main__":
  
  main(sys.argv)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Network definitions for the PPO algorithm."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools
import operator

try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

NetworkOutput = collections.namedtuple('NetworkOutput', 'policy, mean, logstd, value, state')


def feed_forward_gaussian(config, action_size, observations, unused_length, state=None):
  """Independent feed forward networks for policy and value.

  The policy network outputs the mean action and the log standard deviation
  is learned as independent parameter vector.

  Args:
    config: Configuration object.
    action_size: Length of the action vector.
    observations: Sequences of observations.
    unused_length: Batch of sequence lengths.
    state: Batch of initial recurrent states.

  Returns:
    NetworkOutput tuple.
  """
  mean_weights_initializer = tf.contrib.layers.variance_scaling_initializer(
      factor=config.init_mean_factor)
  logstd_initializer = tf.random_normal_initializer(config.init_logstd, 1e-10)
  flat_observations = tf.reshape(observations, [
      tf.shape(observations)[0],
      tf.shape(observations)[1],
      functools.reduce(operator.mul,
                       observations.shape.as_list()[2:], 1)
  ])
  with tf.variable_scope('policy'):
    x = flat_observations
    for size in config.policy_layers:
      x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
    mean = tf.contrib.layers.fully_connected(x,
                                             action_size,
                                             tf.tanh,
                                             weights_initializer=mean_weights_initializer)
    logstd = tf.get_variable('logstd', mean.shape[2:], tf.float32, logstd_initializer)
    logstd = tf.tile(logstd[None, None],
                     [tf.shape(mean)[0], tf.shape(mean)[1]] + [1] * (mean.shape.ndims - 2))
  with tf.variable_scope('value'):
    x = flat_observations
    for size in config.value_layers:
      x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
    value = tf.contrib.layers.fully_connected(x, 1, None)[..., 0]
  mean = tf.check_numerics(mean, 'mean')
  logstd = tf.check_numerics(logstd, 'logstd')
  value = tf.check_numerics(value, 'value')
  policy = tf.contrib.distributions.MultivariateNormalDiag(mean, tf.exp(logstd))
  return NetworkOutput(policy, mean, logstd, value, state)


def recurrent_gaussian(config, action_size, observations, length, state=None):
  """Independent recurrent policy and feed forward value networks.

  The policy network outputs the mean action and the log standard deviation
  is learned as independent parameter vector. The last policy layer is
  recurrent and uses a GRU cell.

  Args:
    config: Configuration object.
    action_size: Length of the action vector.
    observations: Sequences of observations.
    length: Batch of sequence lengths.
    state: Batch of initial recurrent states.

  Returns:
    NetworkOutput tuple.
  """
  mean_weights_initializer = tf.contrib.layers.variance_scaling_initializer(
      factor=config.init_mean_factor)
  logstd_initializer = tf.random_normal_initializer(config.init_logstd, 1e-10)
  cell = tf.contrib.rnn.GRUBlockCell(config.policy_layers[-1])
  flat_observations = tf.reshape(observations, [
      tf.shape(observations)[0],
      tf.shape(observations)[1],
      functools.reduce(operator.mul,
                       observations.shape.as_list()[2:], 1)
  ])
  with tf.variable_scope('policy'):
    x = flat_observations
    for size in config.policy_layers[:-1]:
      x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
    x, state = tf.nn.dynamic_rnn(cell, x, length, state, tf.float32)
    mean = tf.contrib.layers.fully_connected(x,
                                             action_size,
                                             tf.tanh,
                                             weights_initializer=mean_weights_initializer)
    logstd = tf.get_variable('logstd', mean.shape[2:], tf.float32, logstd_initializer)
    logstd = tf.tile(logstd[None, None],
                     [tf.shape(mean)[0], tf.shape(mean)[1]] + [1] * (mean.shape.ndims - 2))
  with tf.variable_scope('value'):
    x = flat_observations
    for size in config.value_layers:
      x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
    value = tf.contrib.layers.fully_connected(x, 1, None)[..., 0]
  mean = tf.check_numerics(mean, 'mean')
  logstd = tf.check_numerics(logstd, 'logstd')
  value = tf.check_numerics(value, 'value')
  policy = tf.contrib.distributions.MultivariateNormalDiag(mean, tf.exp(logstd))
  # assert state.shape.as_list()[0] is not None
  return NetworkOutput(policy, mean, logstd, value, state)



import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
print("current_dir=" + currentdir)
os.sys.path.insert(0, currentdir)

import math
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import time
import pybullet as p
from . import kuka
import random
import pybullet_data
from pkg_resources import parse_version

largeValObservation = 100

RENDER_HEIGHT = 720
RENDER_WIDTH = 960


class KukaGymEnv(gym.Env):
  metadata = {'render.modes': ['human', 'rgb_array'], 'video.frames_per_second': 50}

  def __init__(self,
               urdfRoot=pybullet_data.getDataPath(),
               actionRepeat=1,
               isEnableSelfCollision=True,
               renders=False,
               isDiscrete=False,
               maxSteps=1000):
    #print("KukaGymEnv __init__")
    self._isDiscrete = isDiscrete
    self._timeStep = 1. / 240.
    self._urdfRoot = urdfRoot
    self._actionRepeat = actionRepeat
    self._isEnableSelfCollision = isEnableSelfCollision
    self._observation = []
    self._envStepCounter = 0
    self._renders = renders
    self._maxSteps = maxSteps
    self.terminated = 0
    self._cam_dist = 1.3
    self._cam_yaw = 180
    self._cam_pitch = -40

    self._p = p
    if self._renders:
      cid = p.connect(p.SHARED_MEMORY)
      if (cid < 0):
        cid = p.connect(p.GUI)
      p.resetDebugVisualizerCamera(1.3, 180, -41, [0.52, -0.2, -0.33])
    else:
      p.connect(p.DIRECT)
    #timinglog = p.startStateLogging(p.STATE_LOGGING_PROFILE_TIMINGS, "kukaTimings.json")
    self.seed()
    self.reset()
    observationDim = len(self.getExtendedObservation())
    #print("observationDim")
    #print(observationDim)

    observation_high = np.array([largeValObservation] * observationDim)
    if (self._isDiscrete):
      self.action_space = spaces.Discrete(7)
    else:
      action_dim = 3
      self._action_bound = 1
      action_high = np.array([self._action_bound] * action_dim)
      self.action_space = spaces.Box(-action_high, action_high)
    self.observation_space = spaces.Box(-observation_high, observation_high)
    self.viewer = None

  def reset(self):
    #print("KukaGymEnv _reset")
    self.terminated = 0
    p.resetSimulation()
    p.setPhysicsEngineParameter(numSolverIterations=150)
    p.setTimeStep(self._timeStep)
    p.loadURDF(os.path.join(self._urdfRoot, "plane.urdf"), [0, 0, -1])

    p.loadURDF(os.path.join(self._urdfRoot, "table/table.urdf"), 0.5000000, 0.00000, -.820000,
               0.000000, 0.000000, 0.0, 1.0)

    xpos = 0.55 + 0.12 * random.random()
    ypos = 0 + 0.2 * random.random()
    ang = 3.14 * 0.5 + 3.1415925438 * random.random()
    orn = p.getQuaternionFromEuler([0, 0, ang])
    self.blockUid = p.loadURDF(os.path.join(self._urdfRoot, "block.urdf"), xpos, ypos, -0.15,
                               orn[0], orn[1], orn[2], orn[3])

    p.setGravity(0, 0, -10)
    self._kuka = kuka.Kuka(urdfRootPath=self._urdfRoot, timeStep=self._timeStep)
    self._envStepCounter = 0
    p.stepSimulation()
    self._observation = self.getExtendedObservation()
    return np.array(self._observation)

  def __del__(self):
    p.disconnect()

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def getExtendedObservation(self):
    self._observation = self._kuka.getObservation()
    gripperState = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaGripperIndex)
    gripperPos = gripperState[0]
    gripperOrn = gripperState[1]
    blockPos, blockOrn = p.getBasePositionAndOrientation(self.blockUid)

    invGripperPos, invGripperOrn = p.invertTransform(gripperPos, gripperOrn)
    gripperMat = p.getMatrixFromQuaternion(gripperOrn)
    dir0 = [gripperMat[0], gripperMat[3], gripperMat[6]]
    dir1 = [gripperMat[1], gripperMat[4], gripperMat[7]]
    dir2 = [gripperMat[2], gripperMat[5], gripperMat[8]]

    gripperEul = p.getEulerFromQuaternion(gripperOrn)
    #print("gripperEul")
    #print(gripperEul)
    blockPosInGripper, blockOrnInGripper = p.multiplyTransforms(invGripperPos, invGripperOrn,
                                                                blockPos, blockOrn)
    projectedBlockPos2D = [blockPosInGripper[0], blockPosInGripper[1]]
    blockEulerInGripper = p.getEulerFromQuaternion(blockOrnInGripper)
    #print("projectedBlockPos2D")
    #print(projectedBlockPos2D)
    #print("blockEulerInGripper")
    #print(blockEulerInGripper)

    #we return the relative x,y position and euler angle of block in gripper space
    blockInGripperPosXYEulZ = [blockPosInGripper[0], blockPosInGripper[1], blockEulerInGripper[2]]

    #p.addUserDebugLine(gripperPos,[gripperPos[0]+dir0[0],gripperPos[1]+dir0[1],gripperPos[2]+dir0[2]],[1,0,0],lifeTime=1)
    #p.addUserDebugLine(gripperPos,[gripperPos[0]+dir1[0],gripperPos[1]+dir1[1],gripperPos[2]+dir1[2]],[0,1,0],lifeTime=1)
    #p.addUserDebugLine(gripperPos,[gripperPos[0]+dir2[0],gripperPos[1]+dir2[1],gripperPos[2]+dir2[2]],[0,0,1],lifeTime=1)

    self._observation.extend(list(blockInGripperPosXYEulZ))
    return self._observation

  def step(self, action):
    if (self._isDiscrete):
      dv = 0.005
      dx = [0, -dv, dv, 0, 0, 0, 0][action]
      dy = [0, 0, 0, -dv, dv, 0, 0][action]
      da = [0, 0, 0, 0, 0, -0.05, 0.05][action]
      f = 0.3
      realAction = [dx, dy, -0.002, da, f]
    else:
      #print("action[0]=", str(action[0]))
      dv = 0.005
      dx = action[0] * dv
      dy = action[1] * dv
      da = action[2] * 0.05
      f = 0.3
      realAction = [dx, dy, -0.002, da, f]
    return self.step2(realAction)

  def step2(self, action):
    for i in range(self._actionRepeat):
      self._kuka.applyAction(action)
      p.stepSimulation()
      if self._termination():
        break
      self._envStepCounter += 1
    if self._renders:
      time.sleep(self._timeStep)
    self._observation = self.getExtendedObservation()

    #print("self._envStepCounter")
    #print(self._envStepCounter)

    done = self._termination()
    npaction = np.array([
        action[3]
    ])  #only penalize rotation until learning works well [action[0],action[1],action[3]])
    actionCost = np.linalg.norm(npaction) * 10.
    #print("actionCost")
    #print(actionCost)
    reward = self._reward() - actionCost
    #print("reward")
    #print(reward)

    #print("len=%r" % len(self._observation))

    return np.array(self._observation), reward, done, {}

  def render(self, mode="rgb_array", close=False):
    if mode != "rgb_array":
      return np.array([])

    base_pos, orn = self._p.getBasePositionAndOrientation(self._kuka.kukaUid)
    view_matrix = self._p.computeViewMatrixFromYawPitchRoll(cameraTargetPosition=base_pos,
                                                            distance=self._cam_dist,
                                                            yaw=self._cam_yaw,
                                                            pitch=self._cam_pitch,
                                                            roll=0,
                                                            upAxisIndex=2)
    proj_matrix = self._p.computeProjectionMatrixFOV(fov=60,
                                                     aspect=float(RENDER_WIDTH) / RENDER_HEIGHT,
                                                     nearVal=0.1,
                                                     farVal=100.0)
    (_, _, px, _, _) = self._p.getCameraImage(width=RENDER_WIDTH,
                                              height=RENDER_HEIGHT,
                                              viewMatrix=view_matrix,
                                              projectionMatrix=proj_matrix,
                                              renderer=self._p.ER_BULLET_HARDWARE_OPENGL)
    #renderer=self._p.ER_TINY_RENDERER)

    rgb_array = np.array(px, dtype=np.uint8)
    rgb_array = np.reshape(rgb_array, (RENDER_HEIGHT, RENDER_WIDTH, 4))

    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def _termination(self):
    #print (self._kuka.endEffectorPos[2])
    state = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaEndEffectorIndex)
    actualEndEffectorPos = state[0]

    #print("self._envStepCounter")
    #print(self._envStepCounter)
    if (self.terminated or self._envStepCounter > self._maxSteps):
      self._observation = self.getExtendedObservation()
      return True
    maxDist = 0.005
    closestPoints = p.getClosestPoints(self._kuka.trayUid, self._kuka.kukaUid, maxDist)

    if (len(closestPoints)):  #(actualEndEffectorPos[2] <= -0.43):
      self.terminated = 1

      #print("terminating, closing gripper, attempting grasp")
      #start grasp and terminate
      fingerAngle = 0.3
      for i in range(100):
        graspAction = [0, 0, 0.0001, 0, fingerAngle]
        self._kuka.applyAction(graspAction)
        p.stepSimulation()
        fingerAngle = fingerAngle - (0.3 / 100.)
        if (fingerAngle < 0):
          fingerAngle = 0

      for i in range(1000):
        graspAction = [0, 0, 0.001, 0, fingerAngle]
        self._kuka.applyAction(graspAction)
        p.stepSimulation()
        blockPos, blockOrn = p.getBasePositionAndOrientation(self.blockUid)
        if (blockPos[2] > 0.23):
          #print("BLOCKPOS!")
          #print(blockPos[2])
          break
        state = p.getLinkState(self._kuka.kukaUid, self._kuka.kukaEndEffectorIndex)
        actualEndEffectorPos = state[0]
        if (actualEndEffectorPos[2] > 0.5):
          break

      self._observation = self.getExtendedObservation()
      return True
    return False

  def _reward(self):

    #rewards is height of target object
    blockPos, blockOrn = p.getBasePositionAndOrientation(self.blockUid)
    closestPoints = p.getClosestPoints(self.blockUid, self._kuka.kukaUid, 1000, -1,
                                       self._kuka.kukaEndEffectorIndex)

    reward = -1000

    numPt = len(closestPoints)
    #print(numPt)
    if (numPt > 0):
      #print("reward:")
      reward = -closestPoints[0][8] * 10
    if (blockPos[2] > 0.2):
      reward = reward + 10000
      print("successfully grasped a block!!!")
      #print("self._envStepCounter")
      #print(self._envStepCounter)
      #print("self._envStepCounter")
      #print(self._envStepCounter)
      #print("reward")
      #print(reward)
    #print("reward")
    #print(reward)
    return reward

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



"""This file implements the functionalities of an example simulated system using pybullet.

"""
import copy
import math
import numpy as np
import os
import pybullet_data
import random


class BoxStackPyBulletSim(object):
  """The ExamplePyBulletSim class that adds some objects to the scene, steps the sim and return a reward.

  """

  def __init__(self, pybullet_client, urdf_root=pybullet_data.getDataPath(), time_step=0.01):
    """Constructs an example simulation and reset it to the initial states.

    Args:
      pybullet_client: The instance of BulletClient to manage different
        simulations.
      urdf_root: The path to the urdf folder.
      time_step: The time step of the simulation.
    """
    self._pybullet_client = pybullet_client
    self._urdf_root = urdf_root
    self.m_actions_taken_since_reset = 0
    self.time_step = time_step
    self.stateId = -1
    self.Reset(reload_urdf=True)

  def Reset(self, reload_urdf=False):
    """Reset the minitaur to its initial states.

    Args:
      reload_urdf: Whether to reload the urdf file. If not, Reset() just place
        the minitaur back to its starting position.
    """
    self.m_actions_taken_since_reset = 0
    xPosRange = 0.025
    yPosRange = 0.025
    boxHalfExtents = 0.025

    if reload_urdf:
      camInfo = self._pybullet_client.getDebugVisualizerCamera()
      cameraDistance = camInfo[10]
      print("cameraYaw=", camInfo[8])
      print("cameraPitch=", camInfo[9])
      print("camtarget=", camInfo[11])
      print("projectionMatrix=", camInfo[3])
      self._pybullet_client.resetDebugVisualizerCamera(cameraDistance=0.3,
                                                       cameraYaw=camInfo[8],
                                                       cameraPitch=camInfo[9],
                                                       cameraTargetPosition=camInfo[11])

      plane = self._pybullet_client.loadURDF("plane.urdf")
      texUid = self._pybullet_client.loadTexture("checker_blue.png")
      self._pybullet_client.changeVisualShape(plane, -1, textureUniqueId=texUid)

      self._numObjects = 4  #random number?

      self._cubes = []

      red = [0.97, 0.25, 0.25, 1]
      green = [0.41, 0.68, 0.31, 1]
      yellow = [0.92, 0.73, 0, 1]
      blue = [0, 0.55, 0.81, 1]
      colors = [red, green, yellow, blue]

      for i in range(self._numObjects):
        pos = [0, 0, boxHalfExtents + i * 2 * boxHalfExtents]
        orn = self._pybullet_client.getQuaternionFromEuler([0, 0, 0])
        orn = [0, 0, 0, 1]
        cube = self._pybullet_client.loadURDF("cube_small.urdf", pos, orn)
        self._pybullet_client.changeVisualShape(cube, -1, rgbaColor=colors[i])
        self._cubes.append(cube)

      self._pybullet_client.setGravity(0, 0, -10)
      self.stateId = self._pybullet_client.saveState()
    else:
      if (self.stateId >= 0):
        self._pybullet_client.restoreState(self.stateId)
    index = 0
    for i in self._cubes:
      posX = random.uniform(-xPosRange, xPosRange)
      posY = random.uniform(-yPosRange, yPosRange)
      yaw = random.uniform(-math.pi, math.pi)
      pos = [posX, posY, boxHalfExtents + index * 2 * boxHalfExtents]
      index += 1
      orn = self._pybullet_client.getQuaternionFromEuler([0, 0, yaw])
      self._pybullet_client.resetBasePositionAndOrientation(i, pos, orn)

  def GetActionDimension(self):
    """Get the length of the action list.

    Returns:
      The length of the action list.
    """
    return 4  #self.num_motors

  def GetObservationUpperBound(self):
    """Get the upper bound of the observation.

    Returns:
      The upper bound of an observation. See GetObservation() for the details
        of each element of an observation.
    """
    upper_bound = np.array([0.0] * self.GetObservationDimension())
    upper_bound[0:] = 1.0  # Quaternion of base orientation.
    return upper_bound

  def GetObservationLowerBound(self):
    """Get the lower bound of the observation."""
    return -self.GetObservationUpperBound()

  def GetObservationDimension(self):
    """Get the length of the observation list.

    Returns:
      The length of the observation list.
    """
    sz = len(self.GetObservation())
    print("sz=", sz)
    return sz

  def GetObservation(self):
    """Get the observations of minitaur.

    Returns:
      The observation list. observation[0:4] is the base orientation in quaternion form.
    """
    observation = []
    for i in self._cubes:
      pos, orn = self._pybullet_client.getBasePositionAndOrientation(i)
      observation.extend(list(pos))
      observation.extend(list(orn))
    return observation

  def ApplyAction(self, action):
    """Set the desired action.
    """
    self.m_actions_taken_since_reset += 1

  def Termination(self):
    return self.m_actions_taken_since_reset >= 200


def CreateSim(pybullet_client, urdf_root, time_step):
  sim = BoxStackPyBulletSim(pybullet_client=pybullet_client,
                            urdf_root=urdf_root,
                            time_step=time_step)
  return sim



"""An example to run minitaur gym environment with randomized terrain.

"""

import math

import numpy as np
import tf.compat.v1 as tf
#from google3.pyglib import app
#from google3.pyglib import flags
from pybullet_envs.minitaur.envs import minitaur_randomize_terrain_gym_env

FLAGS = flags.FLAGS

flags.DEFINE_enum("example_name", "reset", ["sine", "reset"],
                  "The name of the example: sine or reset.")


def ResetTerrainExample():
  """An example showing resetting random terrain env."""
  num_reset = 10
  steps = 100
  env = minitaur_randomize_terrain_gym_env.MinitaurRandomizeTerrainGymEnv(
      render=True, leg_model_enabled=False, motor_velocity_limit=np.inf, pd_control_enabled=True)
  action = [math.pi / 2] * 8
  for _ in xrange(num_reset):
    env.reset()
    for _ in xrange(steps):
      _, _, done, _ = env.step(action)
      if done:
        break


def SinePolicyExample():
  """An example of minitaur walking with a sine gait."""
  env = minitaur_randomize_terrain_gym_env.MinitaurRandomizeTerrainGymEnv(
      render=True, motor_velocity_limit=np.inf, pd_control_enabled=True, on_rack=False)
  sum_reward = 0
  steps = 200
  amplitude_1_bound = 0.5
  amplitude_2_bound = 0.5
  speed = 40

  for step_counter in xrange(steps):
    time_step = 0.01
    t = step_counter * time_step

    amplitude1 = amplitude_1_bound
    amplitude2 = amplitude_2_bound
    steering_amplitude = 0
    if t < 10:
      steering_amplitude = 0.5
    elif t < 20:
      steering_amplitude = -0.5
    else:
      steering_amplitude = 0

    # Applying asymmetrical sine gaits to different legs can steer the minitaur.
    a1 = math.sin(t * speed) * (amplitude1 + steering_amplitude)
    a2 = math.sin(t * speed + math.pi) * (amplitude1 - steering_amplitude)
    a3 = math.sin(t * speed) * amplitude2
    a4 = math.sin(t * speed + math.pi) * amplitude2
    action = [a1, a2, a2, a1, a3, a4, a4, a3]
    _, reward, _, _ = env.step(action)
    sum_reward += reward


def main(unused_argv):
  if FLAGS.example_name == "sine":
    SinePolicyExample()
  elif FLAGS.example_name == "reset":
    ResetTerrainExample()


if __name__ == "__main__":
  tf.logging.set_verbosity(tf.logging.INFO)
  app.run()



"""A dummy gait generator module for storing gait patterns from higher-level controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
from typing import Any, Sequence

import gin

from pybullet_envs.minitaur.agents.baseline_controller import gait_generator

LAIKAGO_STANDING = (
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
)


@gin.configurable
class DummyGaitGenerator(gait_generator.GaitGenerator):
  """A module for storing quadruped gait patterns from high-level controller.

  This module stores the state for each leg of a quadruped robot. The data is
  used by the stance leg controller to determine the appropriate contact forces.
  A high-level controller, such as a neural network policy, can be used to
  control the gait pattern and set corresponding leg states.
  """

  def __init__(
      self,
      robot: Any,
      initial_leg_state: Sequence[gait_generator.LegState] = LAIKAGO_STANDING,
  ):
    """Initializes the class.

    Args:
      robot: A quadruped robot.
      initial_leg_state: The desired initial swing/stance state of legs indexed
        by their id.
    """
    self._robot = robot
    if len(initial_leg_state) != len(
        list(self._robot.urdf_loader.get_end_effector_id_dict().values())):
      raise ValueError(
          "The number of leg states should be the same of number of legs.")
    self._initial_leg_state = initial_leg_state
    self._leg_state = list(initial_leg_state)
    self._desired_leg_state = list(initial_leg_state)

    self.reset(0)

  def reset(self, current_time):
    del current_time
    self._leg_state = list(self._initial_leg_state)
    self._desired_leg_state = list(self._initial_leg_state)

  @property
  def desired_leg_state(self) -> Sequence[gait_generator.LegState]:
    """The desired leg SWING/STANCE states.

    Returns:
      The SWING/STANCE states for all legs.

    """
    return self._desired_leg_state

  @desired_leg_state.setter
  def desired_leg_state(self, state):
    self._desired_leg_state = copy.deepcopy(state)

  @property
  def leg_state(self) -> Sequence[gait_generator.LegState]:
    """The leg state after considering contact with ground.

    Returns:
      The actual state of each leg after accounting for contacts.
    """
    return self._leg_state

  @leg_state.setter
  def leg_state(self, state):
    self._leg_state = copy.deepcopy(state)




import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
fileIO = p.loadPlugin("fileIOPlugin")
if (fileIO >= 0):
  #we can have a zipfile (pickup.zip) inside a zipfile (pickup2.zip)
  p.executePluginCommand(fileIO, pybullet_data.getDataPath()+"/pickup2.zip", [p.AddFileIOAction, p.ZipFileIO])
  p.executePluginCommand(fileIO, "pickup.zip", [p.AddFileIOAction, p.ZipFileIO])
  objs = p.loadSDF("pickup/model.sdf")
  dobot = objs[0]
  p.changeVisualShape(dobot, -1, rgbaColor=[1, 1, 1, 1])

else:
  print("fileIOPlugin is disabled.")

p.setPhysicsEngineParameter(enableFileCaching=False)

while (1):
  p.stepSimulation()
  time.sleep(1. / 240.)



"""Trajectory Generator generates walking leg motions for a quadruped robot.

Trajectory Generator (TG) has an internal state (phase) and generates
walking-like motion for 8 motors of minitaur quadruped robot based on
parameters
such as:
 - delta time to progress the TG's internal state.
 - intensity to control amount of movement (stride length and lift of the legs).
 - waking height to control the average extension of the legs.

Each time step() is called, the internal state is progressed and 8 motor
positions are generated. This TG uses the open-loop SineController class to
provide leg positions. It is mainly a wrapper for ability to modulating the
time
and other parameters of the SineController.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import numpy as np
import gin
from pybullet_envs.minitaur.agents.trajectory_generator import controller_simple

PHASE_LOWER_BOUND = 0.0
PHASE_UPPER_BOUND = 1.0
WALK_HEIGHT_LOWER_BOUND = -0.5
WALK_HEIGHT_UPPER_BOUND = 1.0
INTENSITY_LOWER_BOUND = 0.0
INTENSITY_UPPER_BOUND = 1.5
_SWING_STANCE_LOWER_BOUND = 0.2
_SWING_STANCE_UPPER_BOUND = 5.0
_DELTA_SWING_STANCE_CAP = 0.4
_TWO_PI = math.pi * 2.0
_LEG_COUPLING_DICT = {
    "null": [],
    # All the legs are coupled.
    "all coupled": [0, 0, 0, 0],
    # Front legs and back legs are coupled separately.
    "front back": [0, 1, 0, 1],
    # Left legs and right legs are coupled separately.
    "left right": [0, 0, 1, 1],
    # Diagonal legs are coupled (i.e. trottting).
    "diagonal": [0, 1, 1, 0],
    # Each leg is indepenent.
    "decoupled": [0, 1, 2, 3]
}


@gin.configurable
class TgSimple(object):
  """TgSimple class is a simplified trajectory generator for quadruped walking.

  It returns 8 actions for quadruped slow walking behavior
  based on the parameters provided such as intensity, walking height and delta
  time. It returns its internal phase as information.
  """

  def __init__(self,
               walk_height_lower_bound=WALK_HEIGHT_LOWER_BOUND,
               walk_height_upper_bound=WALK_HEIGHT_UPPER_BOUND,
               intensity_lower_bound=INTENSITY_LOWER_BOUND,
               intensity_upper_bound=INTENSITY_UPPER_BOUND,
               swing_stance_lower_bound=_SWING_STANCE_LOWER_BOUND,
               swing_stance_upper_bound=_SWING_STANCE_UPPER_BOUND,
               integrator_coupling_mode="all coupled",
               walk_height_coupling_mode="all coupled",
               variable_swing_stance_ratio=False,
               swing_stance_ratio=1.0,
               init_leg_phase_offsets=None):
    """Initialize the trajectory generator with a controller.

    For trajectory generator, we create an asymmetric sine controller with
    parameters that was previously optimized as an open-loop controller.

    Args:
      walk_height_lower_bound: Lower bound for walking height which sets the
        default leg extension of the gait. Unit is rad, -0.5 by default.
      walk_height_upper_bound: Lower bound for walking height which sets the
        default leg extension of the gait. Unit is rad, 1.0 by default.
      intensity_lower_bound: The upper bound for intensity of the trajectory
        generator. It can be used to limit the leg movement.
      intensity_upper_bound: The upper bound for intensity of the trajectory
        generator. It can be used to limit the leg movement.
      swing_stance_lower_bound: Lower bound for the swing vs stance ratio
        parameter. Default value is 0.2.
      swing_stance_upper_bound: Upper bound for the swing vs stance ratio
        parameter. Default value is 0.2.
      integrator_coupling_mode: How the legs should be coupled for integrators.
      walk_height_coupling_mode: The same coupling mode used for walking
        heights for the legs.
      variable_swing_stance_ratio: A boolean to indicate if the swing stance
        ratio can change per time step or not.
      swing_stance_ratio: Time taken by swing phase vs stance phase. This is
        only relevant if variable_swing_stance_ratio is False.
      init_leg_phase_offsets: The initial phases of the legs. A list of 4
        variables within [0,1). The order is front-left, rear-left, front-right
        and rear-right.

    Raises:
      ValueError: If parameters are not valid values.
    """
    self._walk_height_lower_bound = walk_height_lower_bound
    self._walk_height_upper_bound = walk_height_upper_bound
    self._intensity_lower_bound = intensity_lower_bound
    self._intensity_upper_bound = intensity_upper_bound
    self._swing_stance_lower_bound = swing_stance_lower_bound
    self._swing_stance_upper_bound = swing_stance_upper_bound
    if not init_leg_phase_offsets:
      init_leg_phase_offsets = [0, 0.25, 0.5, 0.75]
    if len(init_leg_phase_offsets) != 4:
      raise ValueError("The number of leg phase offsets is not equal to 4.")
    if min(init_leg_phase_offsets) < 0 or max(init_leg_phase_offsets) >= 1:
      raise ValueError("Leg phase offsets are not within [0,1)")
    self._legs = []
    for period in init_leg_phase_offsets:
      init_phase = period * 2 * math.pi
      self._legs.append(controller_simple.SimpleLegController(init_phase))

    if integrator_coupling_mode not in _LEG_COUPLING_DICT:
      raise ValueError("Invalid integrator_coupling_mode.")
    if walk_height_coupling_mode not in _LEG_COUPLING_DICT:
      raise ValueError("Invalid walk_height_coupling_mode.")

    # Set the phase couplings and build a list of legs per phase coupling.
    self._integrator_id_per_leg = _LEG_COUPLING_DICT[integrator_coupling_mode]
    self._num_integrators = max(
        self._integrator_id_per_leg) + 1 if self._integrator_id_per_leg else 0
    self._legs_per_integrator_id = [[], [], [], []]
    for idx, phase_id in enumerate(self._integrator_id_per_leg):
      self._legs_per_integrator_id[phase_id].append(self._legs[idx])

    # For each integrator coupling, create a integrator unit.
    # For each leg controlled by that phase generator, mark the phase offset.
    self._integrator_units = []
    for legs_per_integrator in self._legs_per_integrator_id:
      if legs_per_integrator:
        circular_integrator = CircularAsymmetricalIntegratorUnit(
            legs_per_integrator[0].phase)
        self._integrator_units.append(circular_integrator)
        for leg in legs_per_integrator:
          leg.phase_offset = leg.phase - circular_integrator.phase

    # Set the walking heights couplings.
    self._walk_height_id_per_leg = _LEG_COUPLING_DICT[walk_height_coupling_mode]
    self._num_walk_heights = max(
        self._walk_height_id_per_leg) + 1 if self._walk_height_id_per_leg else 0
    self._variable_swing_stance_ratio = variable_swing_stance_ratio
    self._swing_stance_ratio = swing_stance_ratio

  def reset(self):
    """Resets leg phase offsets to their initial values."""
    for leg in self._legs:
      leg.reset()
    for circular_integrator in self._integrator_units:
      circular_integrator.reset()

  def get_parameter_bounds(self):
    """Lower and upper bounds for the parameters generator's parameters.

    Returns:
      2-tuple of:
      - Lower bounds for the parameters such as intensity, walking height and
      lift fraction.
      - Upper bounds for the same parameters.
    """
    lower_bounds = [self._intensity_lower_bound]
    upper_bounds = [self._intensity_upper_bound]
    lower_bounds += [self._walk_height_lower_bound] * self._num_walk_heights
    upper_bounds += [self._walk_height_upper_bound] * self._num_walk_heights
    lower_bounds += [self._swing_stance_lower_bound
                    ] * self._variable_swing_stance_ratio
    upper_bounds += [self._swing_stance_upper_bound
                    ] * self._variable_swing_stance_ratio

    return lower_bounds, upper_bounds

  def get_actions(self, delta_real_time, tg_params):
    """Get actions for 8 motors after increasing the phase delta_time.

    Args:
      delta_real_time: Time in seconds that have actually passed since the last
        step of the trajectory generator.
      tg_params: An ndarray of the parameters for generating the trajectory. The
        parameters must be in the correct order (time_scale, intensity,
        walking_height, and swing vs stance)

    Raises:
      ValueError: In case the input dimension does not match expected.
    Returns:
      The rotations for all the 8 motors for this time step
      returned in an array [front_left_motor_1, front_left_motor_2, etc].
    """
    speeds, intensity, heights, swing_stance_ratio = self._process_tg_params(
        tg_params)
    # Adjust the swing stance ratio of the controller (used for all four legs).
    if swing_stance_ratio:
      self.adjust_swing_stance_ratio(swing_stance_ratio)
    # Adjust the walking height, intensity and swing vs stance of the legs.
    for idx, leg in enumerate(self._legs):
      leg.adjust_intensity(intensity)
      if heights:
        leg.adjust_center_extension(heights[self._walk_height_id_per_leg[idx]])

    # Progress all the phase generators based on delta time.
    for idx, integrator_unit in enumerate(self._integrator_units):
      integrator_unit.progress_phase(speeds[idx] * delta_real_time,
                                     self._swing_stance_ratio)

    # Set the phases for the legs based on their offsets with phase generators.
    for phase_id, leg_list in enumerate(self._legs_per_integrator_id):
      for leg in leg_list:
        delta_period = leg.phase_offset / (2.0 * math.pi)
        leg.phase = self._integrator_units[phase_id].calculate_progressed_phase(
            delta_period, self._swing_stance_ratio)

    # Calculate swingextend and convert it to the motor rotations.
    actions = []
    for idx, leg in enumerate(self._legs):
      swing, extend = leg.get_swing_extend()
      actions.extend([swing, extend])
    return actions

  def _process_tg_params(self, tg_params):
    """Process the trajectory generator parameters and split them.

    Args:
      tg_params: A list consisting of time_scales, intensity, walking_heights,
        swing_stance_ratio. The size depends on the configuration and inital
        flags.

    Returns:
      time_scales: A list of multipliers of delta time (one per integrator).
      intensity: Intensity of the trajectory generator (one variable).
      walking_heights: Walking heights used for the legs. The length depends on
        the coupling between the legs selected at the initialization.
      swing_stance_ratio: The ratio of the speed of the leg during swing stance
      vs stance phase.
    """

    # Check if the given input's dimension matches the expectation considering
    # the number of parameters the trajectory generator uses.
    if isinstance(tg_params, np.ndarray):
      tg_params = tg_params.tolist()
    expected_action_dim = 1 + self._num_integrators + self._num_walk_heights
    if self._variable_swing_stance_ratio:
      expected_action_dim += 1
    if len(tg_params) != expected_action_dim:
      raise ValueError(
          "Action dimension does not match the expectation {} vs {}".format(
              len(tg_params), expected_action_dim))
    # Split input into different parts based on type. The order must match the
    # order given by the order in get_parameter_bounds
    time_scales = tg_params[0:self._num_integrators]
    intensity = tg_params[self._num_integrators]
    walking_heights = tg_params[(self._num_integrators + 1):(
        1 + self._num_integrators + self._num_walk_heights)]
    swing_stance_ratio = None
    if self._variable_swing_stance_ratio:
      swing_stance_ratio = tg_params[1 + self._num_integrators +
                                     self._num_walk_heights]

    return time_scales, intensity, walking_heights, swing_stance_ratio

  def get_state(self):
    """Returns a list of floats representing the phase of the controller.

    The phase of the controller is composed of the phases of the integrators.
    For each integrator, the phase is composed of 2 floats that represents the
    sine and cosine of the phase of that integrator.

    Returns:
      List containing sine and cosine of the phases of all the integrators.
    """
    return [x for y in self._integrator_units for x in y.get_state()]

  def get_state_lower_bounds(self):
    """Lower bounds for the internal state.

    Returns:
      The list containing the lower bounds.
    """
    return [PHASE_LOWER_BOUND] * 2 * self._num_integrators

  def get_state_upper_bounds(self):
    """Upper bounds for the internal state.

    Returns:
      The list containing the upper bounds.
    """
    return [PHASE_UPPER_BOUND] * 2 * self._num_integrators

  def adjust_swing_stance_ratio(self, target_swing_stance_ratio):
    """Adjust the parameter swing_stance_ratio towards a given target value.

    Args:
      target_swing_stance_ratio: The target value for the ratio between swing
        and stance phases.
    """
    delta = min(_DELTA_SWING_STANCE_CAP,
                target_swing_stance_ratio - self._swing_stance_ratio)
    self._swing_stance_ratio += delta

  @property
  def num_integrators(self):
    """Gets the number of integrators used based on coupling mode."""
    return self._num_integrators


class CircularAsymmetricalIntegratorUnit(object):
  """A circular integrator with asymmetry between first and second half.

  An integrator is a memory unit that accumulates the given parameter at every
  time step.
  A circular integrator is when the integrator cycles within [0,2pi].
  The phase of a circular integrator indicates the accumulated number and it is
  stored as fmod of 2Pi.
  Asymmetrical circular integrator has a further characteristic where it
  distinguishes between the first half of the period vs the second half. It
  allows the integrator to move at different speeds during these two periods.
  From a locomotion perspective these two halves of the period can be considered
  as swing and stance phases. The speed difference is calculated using the
  variable swing_stance_ratio provided at every time step.
  CircularAsymmetricalIntegratorUnit can be used to control one or multiple legs
  depending on the preference. If more than one leg is assigned to a single unit
  the other legs are calculated based on their initial phase difference.
  """

  def __init__(self, init_phase=0):
    self._init_phase = init_phase
    self.reset()

  def reset(self):
    self.phase = self._init_phase

  def calculate_progressed_phase(self, delta_period, swing_stance_speed_ratio):
    """Calculate a hypotethical phase based on the current phase and args.

    This is used to both calculate the new phase, as well as the current phase
    of the other legs with a given offset of delta_period.

    Args:
      delta_period: The fraction of the period to add to the current phase of
        the integrator. If set to 1, the integrator will accomplish one full
        period and return the same phase. The calculated phase will depend on
        the current phase (if it is in first half vs second half) and swing vs
        stance speed ratio.
      swing_stance_speed_ratio: The ratio of the speed of the phase when it is
        in swing (second half) vs stance (first half). Set to 1.0 by default,
        making it symettric, same as a classical integrator.

    Returns:
      The new phase between 0 and 2 * pi.
    """
    stance_speed_coef = (
        swing_stance_speed_ratio + 1) * 0.5 / swing_stance_speed_ratio
    swing_speed_coef = (swing_stance_speed_ratio + 1) * 0.5
    delta_left = delta_period
    new_phase = self.phase
    while delta_left > 0:
      if 0 <= new_phase < math.pi:
        delta_phase_multiplier = stance_speed_coef * _TWO_PI
        new_phase += delta_left * delta_phase_multiplier
        delta_left = 0
        if new_phase < math.pi:
          delta_left = 0
        else:
          delta_left = (new_phase - math.pi) / delta_phase_multiplier
          new_phase = math.pi
      else:
        delta_phase_multiplier = swing_speed_coef * _TWO_PI
        new_phase += delta_left * delta_phase_multiplier
        if math.pi <= new_phase < _TWO_PI:
          delta_left = 0
        else:
          delta_left = (new_phase - _TWO_PI) / delta_phase_multiplier
          new_phase = 0
    return math.fmod(new_phase, _TWO_PI)

  def progress_phase(self, delta_period, swing_stance_ratio):
    """Updates the phase based on the current phase, delta period and ratio.

    Args:
      delta_period: The fraction of the period to add to the current phase of
        the integrator. If set to 1, the integrator will accomplish one full
        period and return the same phase. The calculated phase will depend on
        the current phase (if it is in first half vs second half) and swing vs
        stance speed ratio.
      swing_stance_ratio: The ratio of the speed of the phase when it is in
        swing (second half) vs stance (first half). Set to 1.0 by default,
        making it symettric, same as a classical integrator.
    """
    self.phase = self.calculate_progressed_phase(delta_period,
                                                 swing_stance_ratio)

  def get_state(self):
    """Returns the sin and cos of the phase as state."""
    return [(math.cos(self.phase) + 1) / 2.0, (math.sin(self.phase) + 1) / 2.0]



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for reinforcement learning."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from . import wrappers
from .attr_dict import AttrDict
from .batch_env import BatchEnv
from .count_weights import count_weights
from .in_graph_batch_env import InGraphBatchEnv
from .in_graph_env import InGraphEnv
from .loop import Loop
from .mock_algorithm import MockAlgorithm
from .mock_environment import MockEnvironment
from .simulate import simulate
from .streaming_mean import StreamingMean



# Lint as: python3
"""Interface that specifies tasks."""
# TODO(tingnan): Def. proper task interface - see TODO(b/154635313) in comments.

import abc
from typing import Sequence

import gym

from pybullet_envs.minitaur.envs_v2.sensors import sensor


class Task(metaclass=abc.ABCMeta):
  """Base class for tasks."""

  # TODO(b/154635313): Deprecate this method. Consolidate it into update().
  @abc.abstractmethod
  def reward(self, env: gym.Env) -> float:
    """Returns the reward for the current state of the environment."""

  @abc.abstractmethod
  def reset(self, env: gym.Env) -> None:
    """Resets the task."""

  @abc.abstractmethod
  def update(self, env: gym.Env) -> None:
    """Updates the internal state of the task."""

  # TODO(b/154635313): Deprecate this method. Consolidate it into update().
  @abc.abstractmethod
  def done(self, env: gym.Env) -> bool:
    """Determines whether the task is done."""

  @property
  def sensors(self) -> Sequence[sensor.Sensor]:
    """Returns sensors used by task."""
    return []



# Lint as: python3
"""A sensor for robot-mounted 1D lidar (laser scan)."""

from typing import Any, Iterable, Sequence, Type, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.vision import point_cloud_utils
from pybullet_envs.minitaur.vision import sim_camera

_MODE_TO_NUM_CHANNELS_DICT = {
    sim_camera.CameraMode.DEPTH: 1,
    sim_camera.CameraMode.RGB: 3,
    sim_camera.CameraMode.RGBD: 4,
    sim_camera.CameraMode.POINTCLOUD_WORLD_FRAME: 3,
    sim_camera.CameraMode.POINTCLOUD_ROBOT_FRAME: 3,
}


@gin.configurable
class CameraSensor(sensor.Sensor):
  """A robot-mounted sensor that returns RGBD images.

  Attributes:
    resolution: A 2-tuple (width, height) that represents the resolution of the
      camera.
    fov_degree: A floating point value that represents the field of view of the
      camera in the vertical direction. The unit is degree.
  """

  def __init__(self,
               camera_translation_from_base,
               camera_rotation_from_base,
               parent_link_id=-1,
               camera_mode=sim_camera.CameraMode.DEPTH,
               camera_update_frequency_hz=10,
               camera_stabilized=False,
               fov_degree=60,
               resolution=(32, 32),
               lower_bound: Union[float, Iterable[float]] = 0.0,
               upper_bound: Union[float, Iterable[float]] = 255.0,
               sensor_latency: Union[float, Sequence[float]] = 0.0,
               dtype: Type[Any] = np.float64,
               name="vision"):
    """Initializes the CameraSensor.

    Args:
      camera_translation_from_base: A 3-vector translation from the center of
        the specified link.
      camera_rotation_from_base: A 4-vector quaternion represents the rotation
        of the camera relative to the specified link.
      parent_link_id: The pybullet link id, where the camera is mounted on.
      camera_mode: An enum that specifies the mode that the camera operates. See
        sim_camera.CameraMode for more details.
      camera_update_frequency_hz: The frequency at which the camera will capture
        a frame.
      camera_stabilized: Whether the camera is stabilized. See
        sim_camera.MountedCamera for more details.
      fov_degree: The vertical field of view of the camera (in degree).
      resolution: A 2-tuple that represents the width and the height of the
        camera image.
      lower_bound: The lower bound of values of the camera output. It could be a
        single float or an array of floats with shape (height, width, channels).
      upper_bound: The upper bound of values of the camera output. It could be a
        single float or an array of floats with shape (height, width,
        channels)..
      sensor_latency: See base class.
      dtype: See base class.
      name: The name of the sensor.
    """
    super().__init__(
        name=name,
        sensor_latency=sensor_latency,
        interpolator_fn=sensor.closest_obs_blender)
    self._parent_link_id = parent_link_id
    self._camera_mode = camera_mode
    self._camera_translation_from_base = camera_translation_from_base
    self._camera_rotation_from_base = camera_rotation_from_base
    self.camera_update_frequency_hz = camera_update_frequency_hz
    self._time_interval_every_camera_update = (1.0 /
                                               self.camera_update_frequency_hz)
    self._camera_stabilized = camera_stabilized
    self._fov_degree = fov_degree
    self._resolution = resolution
    num_channels = _MODE_TO_NUM_CHANNELS_DICT[self._camera_mode]
    self._camera = None
    self._camera_image = None
    self._dtype = dtype
    if isinstance(lower_bound, list):
      lower_bound = np.array(lower_bound, dtype=self._dtype)
    else:
      lower_bound = lower_bound * np.ones(
          shape=(resolution[1], resolution[0], num_channels), dtype=self._dtype)
    if isinstance(upper_bound, list):
      upper_bound = np.array(upper_bound, dtype=self._dtype)
    else:
      upper_bound = upper_bound * np.ones(
          shape=(resolution[1], resolution[0], num_channels), dtype=self._dtype)
    self._observation_space = gym.spaces.Box(
        low=lower_bound, high=upper_bound, dtype=self._dtype)
    self._last_camera_image_timestamp = None

  def change_mounting_point(
      self,
      camera_translation_from_link: Sequence[float] = (0, 0, 0),
      camera_rotation_from_link: Sequence[float] = (0, 0, 0, 1),
      parent_link_id: int = -1):
    """Changes mounting point. Must be called before calls to set_robot().

    Args:
      camera_translation_from_link: A 3-vector translation from the center of
        the specified link.
      camera_rotation_from_link: A 4-vector quaternion represents the rotation
        of the camera relative to the specified link.
      parent_link_id: The pybullet link id, where the camera is mounted on.
    """
    self._parent_link_id = parent_link_id
    self._camera_translation_from_base = camera_translation_from_link
    self._camera_rotation_from_base = camera_rotation_from_link

  def set_robot(self, robot):
    super().set_robot(robot)

    self._camera = sim_camera.MountedCamera(
        pybullet_client=robot.pybullet_client,
        body_id=robot.robot_id,
        parent_link_id=self._parent_link_id,
        relative_translation=self._camera_translation_from_base,
        relative_rotation=self._camera_rotation_from_base,
        stabilized=self._camera_stabilized,
        camera_mode=self._camera_mode,
        fov_degree=self._fov_degree,
        resolution=self._resolution)

  def on_reset(self, env):
    self._env = env
    self._last_camera_image_timestamp = None
    super().on_reset(env)

  def _get_original_observation(self):
    if self._last_camera_image_timestamp is None or (
        self._robot.timestamp >= self._last_camera_image_timestamp +
        self._time_interval_every_camera_update):
      self._camera_image = self._camera.render_image().astype(self._dtype)
      self._last_camera_image_timestamp = self._robot.timestamp
    return self._robot.timestamp, self._camera_image

  def project_depth_map_to_point_cloud(self, depth_map, use_world_frame=True):
    """Convert the depth map into a 3D point cloud.

    Args:
      depth_map: A 2D numpy array with shape (height, width) which represents
        the depth map.
      use_world_frame: Whether converts the depth map into a point cloud in the
        world frame. If False, the point cloud is in the robot's local frame. If
        True, the point cloud is in the world frame if the robot's base
        position/orientation can be measured (e.g. in sim, using SLAM or mocap).

    Returns:
      A point cloud represented by a numpy array of shape (height, width, 3).
    """
    point_cloud = point_cloud_utils.distance_map_to_point_cloud(
        np.squeeze(depth_map), self.fov_degree / 180.0 * np.pi,
        depth_map.shape[1], depth_map.shape[0])
    if use_world_frame:
      point_cloud = (
          self._camera.transform_point_cloud_from_camera_to_world_frame(
              point_cloud))
    else:
      point_cloud = (
          self._camera.transform_point_cloud_from_camera_to_robot_frame(
              point_cloud))
    return point_cloud

  @property
  def resolution(self):
    return self._camera.resolution

  @property
  def fov_degree(self):
    return self._camera.fov_degree



import pybullet as p
import pybullet_data as pd
import math
import time
import numpy as np
import xarm_sim

p.connect(p.GUI)
p.setAdditionalSearchPath(pd.getDataPath())

timeStep=1./60.
p.setTimeStep(timeStep)
p.setGravity(0,0,-9.8)
 
xarm = xarm_sim.XArm6Sim(p,[0,0,0])
while (1):
	xarm.step()
	p.stepSimulation()
	time.sleep(timeStep)
	



import pybullet_data as pd
import pybullet_utils as pu
import pybullet
from pybullet_utils import bullet_client as bc
import time

p = bc.BulletClient(connection_mode=pybullet.GUI)
p.setAdditionalSearchPath(pd.getDataPath())
p.loadURDF("plane_transparent.urdf", useMaximalCoordinates=True)
p  #.setPhysicsEngineParameter(numSolverIterations=10, fixedTimeStep=0.01)

p.configureDebugVisualizer(p.COV_ENABLE_PLANAR_REFLECTION, 1)
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0)

y2z = p.getQuaternionFromEuler([0, 0, 1.57])
meshScale = [1, 1, 1]
visualShapeId = p.createVisualShape(shapeType=p.GEOM_MESH,
                                    fileName="domino/domino.obj",
                                    rgbaColor=[1, 1, 1, 1],
                                    specularColor=[0.4, .4, 0],
                                    visualFrameOrientation=y2z,
                                    meshScale=meshScale)

boxDimensions = [0.5 * 0.00635, 0.5 * 0.0254, 0.5 * 0.0508]
collisionShapeId = p.createCollisionShape(p.GEOM_BOX, halfExtents=boxDimensions)

for j in range(12):
  print("j=", j)
  for i in range(35):
    #p.loadURDF("domino/domino.urdf",[i*0.04,0, 0.06])
    p.createMultiBody(baseMass=0.025,
                      baseCollisionShapeIndex=collisionShapeId,
                      baseVisualShapeIndex=visualShapeId,
                      basePosition=[i * 0.04, j * 0.05, 0.06],
                      useMaximalCoordinates=True)

p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)

p.setGravity(0, 0, -9.8)
p.setRealTimeSimulation(1)
while (1):
  p.setGravity(0, 0, -9.8)
  #p.stepSimulation(1./100.)
  time.sleep(1. / 240.)



import pybullet as p
import time
import math

import pybullet_data
p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())

p.loadURDF("plane.urdf")
cubeId = p.loadURDF("cube_small.urdf", 0, 0, 1)
p.setGravity(0, 0, -10)
p.setRealTimeSimulation(1)
cid = p.createConstraint(cubeId, -1, -1, -1, p.JOINT_FIXED, [0, 0, 0], [0, 0, 0], [0, 0, 1])
print(cid)
print(p.getConstraintUniqueId(0))
a = -math.pi
while 1:
  a = a + 0.01
  if (a > math.pi):
    a = -math.pi
  time.sleep(.01)
  p.setGravity(0, 0, -10)
  pivot = [a, 0, 1]
  orn = p.getQuaternionFromEuler([a, 0, 0])
  p.changeConstraint(cid, pivot, jointChildFrameOrientation=orn, maxForce=50)

p.removeConstraint(cid)



# Lint as: python3
"""Crowd objects/human controllers module."""

import abc
import collections
from typing import Any, Callable, Dict, Iterable, List, Optional, Union, Sequence, Text

from absl import logging
import dataclasses
import gin
import numpy as np
#import rvo2

from pybullet_envs.minitaur.envs_v2.sensors import base_position_sensor
from pybullet_envs.minitaur.envs_v2.sensors import sensor as generic_sensor
from pybullet_envs.minitaur.robots import autonomous_object
from pybullet_envs.minitaur.robots import object_controller


POSITION_SENSOR_POSTFIX = "_pos"


@dataclasses.dataclass
class MovingObjectRecord:
  position_key: Text
  agent_id: int
  radius: float
  last_position: Optional[np.ndarray] = None


@gin.configurable
def sample_start_target_position(scene,
                                 start=None,
                                 start_circles=None,
                                 target_circles=None,
                                 num_sampling_retries=1,
                                 min_wall_distance=0.0,
                                 min_goal_euclidean_distance=0.0,
                                 max_goal_euclidean_distance=np.Inf,
                                 min_path_clearance=None):
  """Sample valid start and target position reachable from start.

  Args:
    scene: a SceneBase instance implementing get_random_valid_position function.
    start: a 2-tuple (x, y) of start position. If specified, no start is
      sampled.
    start_circles: a list of circle specification. Each circle is specified as
      a tuple ((x, y), r) of a center (x, y) and radius r. If specified, start
      position is sampled from within one of the start_circles.
    target_circles: same as start_circle. If specified, target positions is
      sampled from within one of the start_circles.
    num_sampling_retries: a positive int, number of attempts to sample a
      start, target pair.
    min_wall_distance: a float, the minimum distance to a wall.
    min_goal_euclidean_distance: a positive float, the minimum distance between
      start and target.
    max_goal_euclidean_distance: a positive float, the maximum distance between
      start and target.
    min_path_clearance: float, clearance of shortest path to walls.

  Returns:
    A 4 tuple (start, target, shortest_path, is_valid). start and target are
    start and target positions, shortest_path is a list of 2-tuples specifying
    the shortest path from start to target, is_valid is bool specifying whether
    the start, target pair is valid. If min_path_clearance is not specified,
    then shortest_path is None.
  """
  if not hasattr(scene, "get_random_valid_position"):
    raise ValueError(
        "Incompatible scene {}. Expected to have `get_random_valid_position` "
        "method.".format(scene))

  def _print_counters(counters):
    for name, value in counters.items():
      logging.info("  %s: %d", name, value)

  sampling_counters = collections.defaultdict(lambda: 0)
  for _ in range(num_sampling_retries):
    if start is None:
      start_pos = scene.get_random_valid_position(
          min_wall_distance, inclusion_circles=start_circles)
    else:
      if start_circles is not None:
        raise ValueError("At most one of the arguments start and start_circles "
                         "can be not None.")
      start_pos = start
    target_pos = scene.get_random_valid_position(
        min_wall_distance, inclusion_circles=target_circles)
    sampling_counters["attempts"] += 1

    euclidean_distance = np.linalg.norm(target_pos - start_pos)
    if euclidean_distance < min_goal_euclidean_distance:
      sampling_counters["min_euclidean"] += 1
      continue
    if euclidean_distance > max_goal_euclidean_distance:
      sampling_counters["max_euclidean"] += 1
      continue

    # Skip the path computation is no path clearance is provided.
    if min_path_clearance is None:
      logging.info("Valid goal with no minimum path clearance checking.")
      _print_counters(sampling_counters)
      return start_pos, target_pos, None, True

    # Check the goal clearance along the shortest path
    if not hasattr(scene, "find_shortest_path"):
      raise ValueError(
          f"scene %s missing find_shortest_path method {scene}")

    # This is a slow process.
    shortest_path = scene.find_shortest_path(
        start_pos[:2], target_pos[:2], min_path_clearance)
    # No path exists between current robot position and goal satisfying the
    # clearance.
    if shortest_path is None:
      sampling_counters["path_clearance"] += 1
      continue

    logging.info("Valid start/target with path clearance checking.")
    _print_counters(sampling_counters)
    return start_pos, target_pos, shortest_path, True

  logging.info("No valid start/target found.")
  _print_counters(sampling_counters)
  return start_pos, target_pos, None, False


class CrowdController(metaclass=abc.ABCMeta):
  """Crowd controller interface."""

  def __init__(self, names: Iterable[Text],
               position_key_formatter="%s" + POSITION_SENSOR_POSTFIX):
    """Constructor.

    Args:
      names: Name of instance (dynamic object or human).
      position_key_formatter: Formatter to convert name to position sensor name.
    """
    self._names = list(names)
    self._position_key_formatter = position_key_formatter
    self._num_instance = len(self._names)

    self._current_time = 0

  def _validate_instance_id(self, instance_id):
    if not 0 <= instance_id < self._num_instance:
      raise ValueError(
          f"instance_id must be an integer in [0, {self.num_instance}), "
          f"got {instance_id}.")

  @property
  def num_instance(self):
    """Returns the number of crowd instances."""
    return self._num_instance

  def instance_name(self, instance_id: int) -> Text:
    """Returns the name of instance."""
    self._validate_instance_id(instance_id)
    return self._names[instance_id]

  def instance_controller(
      self, instance_id: int) -> object_controller.ControllerBase:
    """Returns the individual controller of certain instance."""
    self._validate_instance_id(instance_id)
    return _IndividualController(self, instance_id)

  def instance_get_action(
      self, instance_id: int, time_sec: float,
      observations: Dict[Text, Any]) -> object_controller.ControllerOutput:
    """Returns action of specific instance given observation.

    This method is for _IndividualController.

    Args:
      instance_id: Identifier of an object in the crowd.
      time_sec: Time since simulation reset in seconds. If time < 0, returns
        initial values and ignores observations.
      observations: A dict of all observations.

    Returns:
      Position, orientation and an extra info dict for robot joints, human
        skeletal pose, etc.
    """
    if time_sec < 0:
      self._recalculate_actions(object_controller.INIT_TIME, {})
      self._current_time = object_controller.INIT_TIME
    elif time_sec > self._current_time:
      self._current_time = time_sec
      self._recalculate_actions(self._current_time, observations)

    self._validate_instance_id(instance_id)

    return self._get_action_of_instance(instance_id)

  @abc.abstractmethod
  def _recalculate_actions(
      self, time_sec: float, observations: Dict[Text, Any]) -> None:
    """Calculates crowd command for all instances in crowd."""
    raise NotImplementedError(
        "_recalculate_actions() should be implemented by subclass.")

  @abc.abstractmethod
  def _get_action_of_instance(
      self, instance_id: int) -> object_controller.ControllerOutput:
    """Returns calculated actions of specific instance."""
    raise NotImplementedError(
        "_get_action_of_instance() should be implemented by subclass.")

  def set_scene(self, scene) -> None:
    """Sets the scene for crowd controller to obtain scene information."""
    del scene


class _IndividualController(object_controller.ControllerBase):
  """A utility class that wraps crowd controller in ControllerBase interface."""

  def __init__(self, crowd_controller: CrowdController, instance_id: int):
    """Constructor.

    Args:
      crowd_controller: The controller of crowd to which this instance belong.
      instance_id: Identifier of a crowd instance.
    """
    self._instance_id = instance_id
    self._crowd_controller = crowd_controller

  def get_action(
      self, time_sec: float,
      observations: Dict[Text, Any]) -> object_controller.ControllerOutput:
    """Returns position, orientation and pose based on time and observations.

    Args:
      time_sec: Time since simulation reset in seconds. If time < 0, returns
        initial values and ignores observations.
      observations: A dict of all observations.

    Returns:
      Position, orientation and an extra info dict for robot joints, human
        skeletal pose, etc.
    """
    return self._crowd_controller.instance_get_action(
        self._instance_id, time_sec, observations)


@gin.configurable
class StationaryController(CrowdController):
  """A crowd controller that places crowd objects at fixed positions."""

  def __init__(
      self, positions: Sequence[Sequence[float]],
      orientations: Optional[Sequence[Sequence[float]]] = None, **kwargs):
    """Constructor.

    Args:
      positions: Fixed positions (3D points) of crowd instances.
      orientations: Fixed orientations in quaternion of crowd instances.
      **kwargs: Keyword arguments to pass on to base class.
    """
    super().__init__(**kwargs)

    if orientations is None:
      orientations = np.array(((0, 0, 0, 1),) * self.num_instance)

    if not len(positions) == len(orientations) == self.num_instance:
      raise ValueError(
          f"positions and orientations should all have the same length "
          f"{self.num_instance}. Got len(positions) = {len(positions)}, "
          f"len(orientations) = {len(orientations)}.")

    self._positions = positions
    self._orientations = orientations

  def _recalculate_actions(
      self, time_sec: float, observations: Dict[Text, Any]) -> None:
    """Calculates crowd command for all instances in crowd."""
    del time_sec
    del observations

  def _get_action_of_instance(
      self, instance_id: int) -> object_controller.ControllerOutput:
    """Returns calculated actions of specific instance."""
    self._validate_instance_id(instance_id)
    return self._positions[instance_id], self._orientations[instance_id], {}


@gin.configurable
class OrcaController(CrowdController):
  """A crowd controller that controls crowd instances using ORCA algorithm.

  Crowd instance will be initialized at a specified start position and move
  towards specified target position in a linear path while avoid collision with
  each other.
  """

  _DEFAULT_NEIGHBOR_DISTANCE_M = 5
  _DEFAULT_MAX_NEIGHBORS = 10
  _DEFAULT_RADIUS_M = 0.5
  _DEFAULT_MAX_SPEED_MPS = 2
  _DEFAULT_TIME_HORIZON_SEC = 1.0
  _DEFAULT_OBSTACLE_TIME_HORIZON_SEC = 0.3

  def __init__(
      self,
      timestep: float,
      start_positions: Optional[Sequence[Sequence[float]]] = None,
      target_positions: Optional[Sequence[Sequence[float]]] = None,
      use_position_generator: Optional[bool] = False,
      group_sizes: Sequence[int] = None,
      radius: float = _DEFAULT_RADIUS_M,
      max_speed_mps: float = _DEFAULT_MAX_SPEED_MPS,
      time_horizon_sec: float = _DEFAULT_TIME_HORIZON_SEC,
      obstacle_time_horizon_sec: float = _DEFAULT_OBSTACLE_TIME_HORIZON_SEC,
      neighbor_distance_m: float = _DEFAULT_NEIGHBOR_DISTANCE_M,
      max_neighbors: int = _DEFAULT_MAX_NEIGHBORS,
      workaround_erp_issue: bool = True,
      moving_objects_pos_key: Sequence[Text] = (),
      moving_objects_radius: Union[float, Sequence[float]] = _DEFAULT_RADIUS_M,
      endless_trajectory: bool = True,
      **kwargs):
    """Constructor.

    Args:
      timestep: Timestep of simulation.
      start_positions: A list of position (x, y, z) for crowd instances as
        their starting position.
      target_positions: A list of position (x, y, z) for crowd instances as
        their target position.
      use_position_generator: a boolean, if True than the start and end
        positions are sampled. start_positions and target_positions must be None
      group_sizes: If set, then crowd is split in groups randomly, whose sizes
        are picked in random from this group_size list. In this way, the
        crowd simulator sumulaters clusters of objects moving around.
      radius: Radius of crowd instances.
      max_speed_mps: Maximum crowd instance speed.
      time_horizon_sec: Time horizon in second.
      obstacle_time_horizon_sec: Time horizon for static obstacle in second.
      neighbor_distance_m: Neighbor distance in meters. Instances closer than
        this distance are considered neighbors.
      max_neighbors: Max number of neighbors.
      workaround_erp_issue: There is an issue with pybullet constraint that the
        constraint is solved only 20% per timestep. Need to amplify position
        delta by 5x to workaround this issue.
      moving_objects_pos_key: Position observation key of moving objects not
        controlled by the ORCA controller.
      moving_objects_radius: Radius of moving objects. Should be a float, which
        applies to all moving objects, or a sequence of float, which should be
        of the same length as moving_objects_pos_key.
      endless_trajectory: Only valid if use_position_generator is True. Agent
        returns to starting point after reaching goal to achieve endless motion.
      **kwargs: Keyword arguments to pass on to base class.
    """
    super().__init__(**kwargs)

    assert ((start_positions is not None and target_positions is not None) or
            use_position_generator)
    if not use_position_generator:
      if not len(start_positions) == len(target_positions) == self.num_instance:
        raise ValueError(
            f"start_positions and target_positions should both have length "
            f"equals {self.num_instance}: "
            f"len(start_positions) = {len(start_positions)}, "
            f"len(target_positions) = {len(target_positions)}.")

    self._timestep = timestep
    self._radius = radius
    self._max_speed_mps = max_speed_mps
    self._time_horizon_sec = time_horizon_sec
    self._obstacle_time_horizon_sec = obstacle_time_horizon_sec
    self._neighbor_distance_m = neighbor_distance_m
    self._max_neighbors = max_neighbors
    self._use_position_generator = use_position_generator
    self._endless_trajectory = endless_trajectory
    self._scene = None
    if isinstance(moving_objects_radius, float):
      moving_objects_radius = [
          moving_objects_radius] * len(moving_objects_pos_key)
    if len(moving_objects_radius) != len(moving_objects_pos_key):
      raise ValueError(
          "moving_objects_radius should be either a float or a sequence of "
          "float with the same length as moving_objects_pos_key.")
    self._moving_objects = [
        MovingObjectRecord(position_key=key, agent_id=-1, radius=radius)
        for key, radius in zip(moving_objects_pos_key, moving_objects_radius)]

    self._paths = None
    self._path_indices = None
    if self._use_position_generator:
      self._start_positions = None
      self._target_positions = None
    else:
      self._start_positions = np.array(start_positions, dtype=np.float64)
      self._target_positions = np.array(target_positions, dtype=np.float64)
    # A guard against multiple initializations. See recalculate_actions below.
    self._already_initialized = False
    self._group_sizes = [1] if group_sizes is None else group_sizes

    # The following variables are initialized in _recalculate_actions()
    self._current_positions = None
    self._command_positions = None
    self._command_orientations = None

    #self._orca = rvo2.PyRVOSimulator(
    #    self._timestep,  # timestep
    #    self._neighbor_distance_m,  # neighborDist
    #    self._max_neighbors,  # maxNeighbors
    #    self._time_horizon_sec,  # timeHorizon
    #    self._obstacle_time_horizon_sec,  # timeHorizonObst
    #    self._radius,  # radius
    #    self._max_speed_mps  # maxSpeed
    #)
    for i in range(self.num_instance):
      if self._use_position_generator:
        start_position = (0, 0)
      else:
        start_position = self._start_positions[i, :2]
      agent_id = self._orca.addAgent(
          tuple(start_position),
          self._neighbor_distance_m,  # neighborDist
          self._max_neighbors,  # maxNeighbors
          self._time_horizon_sec,  # timeHorizon
          self._obstacle_time_horizon_sec,  # timeHorizonObst
          self._radius,  # radius
          self._max_speed_mps,  # maxSpeed
          (0.0, 0.0))  # velocity
      assert agent_id == i

    for obj in self._moving_objects:
      obj.agent_id = self._orca.addAgent(
          (0.0, 0.0),  # position (will adjust after simulation starts)
          self._neighbor_distance_m,  # neighborDist
          self._max_neighbors,  # maxNeighbors
          self._timestep,  # timeHorizon
          self._timestep,  # timeHorizonObst
          obj.radius,  # radius
          self._max_speed_mps,  # maxSpeed
          (0.0, 0.0))  # velocity

    self._workaround_erp_issue = workaround_erp_issue

  def _subsample_path(self, path, subsample_step=1.0):
    subsampled_path = [path[0]]
    traveled_dist = 0.0
    for i, (s, t) in enumerate(zip(path[:-1], path[1:])):
      traveled_dist += np.sqrt(
          np.square(s[0] - t[0]) + np.square(s[1] - t[1]))
      if traveled_dist > subsample_step or i >= len(path) - 2:
        subsampled_path.append(t)
        traveled_dist = 0.0
    return subsampled_path

  def _generate_start_target_positions(self):
    """Generates start and target positions using goal generartors."""
    assert self._scene is not None
    self._start_positions = np.zeros((self.num_instance, 3), dtype=np.float64)
    self._target_positions = np.zeros((self.num_instance, 3), dtype=np.float64)

    self._paths = []
    self._path_indices = []
    start_circles, target_circles = None, None
    group_radius = 1.0
    current_group_size = np.random.choice(self._group_sizes)
    index_in_current_group = 0
    for i in range(self._num_instance):
      start_pos, target_pos, path, is_valid = sample_start_target_position(
          self._scene,
          start_circles=start_circles,
          target_circles=target_circles)
      if index_in_current_group == current_group_size - 1:
        start_circles, target_circles = None, None
        index_in_current_group = 0
        current_group_size = np.random.choice(self._group_sizes)
      else:
        if start_circles is None:
          start_circles = [(start_pos[:2], group_radius)]
          target_circles = [(target_pos[:2], group_radius)]
        else:
          start_circles += [(start_pos[:2], group_radius)]
          target_circles += [(target_pos[:2], group_radius)]
        index_in_current_group += 1
      if not is_valid:
        raise ValueError("No valid start/target positions.")
      self._start_positions[i, :] = start_pos
      self._target_positions[i, :] = target_pos

      subsampled_path = self._subsample_path(path)
      self._paths.append(np.array(subsampled_path, dtype=np.float32))
      self._path_indices.append(0)

  def _recalculate_actions(
      self, time_sec: float, observations: Dict[Text, Any]) -> None:
    """Calculates crowd command for all crowd instances."""
    if self._use_position_generator:
      if (time_sec == object_controller.INIT_TIME and
          self._start_positions is None and
          not self._already_initialized):
        self._generate_start_target_positions()
        # Initialize only once per initial time even if recalculate actions
        # is called multiple times.
        self._already_initialized = True
    if time_sec == object_controller.INIT_TIME:
      # Resets orca simulator.
      for i in range(len(self._names)):
        self._orca.setAgentPosition(i, tuple(self._start_positions[i, :2]))

      self._command_positions = self._start_positions.copy()
      self._current_positions = self._start_positions.copy()
      self._command_orientations = np.repeat(
          ((0.0, 0.0, 0.0, 1.0),), len(self._names), axis=0)
      self._last_target_recalculation_sec = time_sec
      return
    else:
      # The moment we step beyond initial time, we can initialize again.
      self._already_initialized = False

    if self._use_position_generator:
      for i in range(self._num_instance):
        dist = np.linalg.norm(
            self._current_positions[i, :] - self._target_positions[i, :])
        if dist < 2.0:
          _, target_pos, path, is_valid = sample_start_target_position(
              self._scene, self._current_positions[i, :])
          if is_valid:
            self._target_positions[i, :] = target_pos
            subsampled_path = self._subsample_path(path)
            self._paths.append(np.array(subsampled_path, dtype=np.float32))
            self._path_indices.append(0)

    # Sets agent position and preferred velocity based on target.
    for i, agent_name in enumerate(self._names):
      position = observations[self._position_key_formatter % agent_name]
      self._orca.setAgentPosition(
          i, tuple(position[:2]))  # ORCA uses 2D position.
      self._current_positions[i, :2] = position[:2]

      if self._paths is not None:
        # Find closest point on the path from start to target, which (1) hasn't
        # been covered already; (2) is at least max_coverage_distance away from
        # current position.
        distances = np.sqrt(np.sum(np.square(
            self._paths[i] - position[:2]), axis=1))
        max_coverage_distance = 1.0
        index = self._path_indices[i]
        while True:
          if index >= len(self._paths[i]) - 1:
            if self._endless_trajectory:
              self._paths[i] = self._paths[i][::-1]
              distances = distances[::-1]
              index = 0
            break
          elif distances[index] > max_coverage_distance:
            break
          else:
            index += 1
        self._path_indices[i] = index
        target_position = self._paths[i][index, :]
      else:
        target_position = self._target_positions[i][:2]

      goal_vector = target_position - position[:2]
      goal_vector_norm = np.linalg.norm(goal_vector) + np.finfo(np.float32).eps
      goal_unit_vector = goal_vector / goal_vector_norm

      kv = 1
      velocity = min(kv * goal_vector_norm,
                     self._DEFAULT_MAX_SPEED_MPS) * goal_unit_vector
      self._orca.setAgentPrefVelocity(i, tuple(velocity))

    for obj in self._moving_objects:
      position = observations[obj.position_key]
      self._orca.setAgentPosition(obj.agent_id, tuple(position[:2]))
      if obj.last_position is None:
        self._orca.setAgentPrefVelocity(obj.agent_id, (0.0, 0.0))
      else:
        velocity = (position - obj.last_position) / self._timestep
        self._orca.setAgentPrefVelocity(obj.agent_id, tuple(velocity[:2]))
      obj.last_position = position.copy()

    # Advances orca simulator.
    self._orca.doStep()

    # Retrieve agent position and save in buffer.
    for i in range(len(self._names)):
      x, y = self._orca.getAgentPosition(i)
      self._command_positions[i, :2] = (x, y)

      yaw = np.arctan2(y - self._current_positions[i, 1],
                       x - self._current_positions[i, 0])
      self._command_orientations[i] = (0, 0, np.sin(yaw / 2), np.cos(yaw / 2))

  def _get_action_of_instance(
      self, instance_id) -> object_controller.ControllerOutput:
    """Returns calculated actions of specific instance."""

    if self._command_positions is None:
      raise RuntimeError(
          "Attempted to get action of instance before _recalculate_actions().")

    self._validate_instance_id(instance_id)

    if self._workaround_erp_issue:
      k_erp = 1 / 0.2
      delta_position = (
          self._command_positions[instance_id] -
          self._current_positions[instance_id])
      command_position = (
          self._current_positions[instance_id] + k_erp * delta_position)
    else:
      command_position = self._command_positions[instance_id].copy()
    return command_position, self._command_orientations[instance_id], {}

  def set_scene(self, scene) -> None:
    """Sets the scene for crowd controller to obtain scene information."""
    try:
      polygons = scene.vectorized_map
      for polygon in polygons:
        self._orca.addObstacle([tuple(point) for point in polygon])
      self._orca.processObstacles()
      self._scene = scene
    except NotImplementedError:
      logging.exception("Scene does not implement vectorized_map property. "
                        "Crowd agent cannot avoid static obstacles.")


@gin.configurable
def uniform_object_factory(
    instance_id: int,
    object_factory: Callable[..., autonomous_object.AutonomousObject],
    *args, **kwargs) -> autonomous_object.AutonomousObject:
  """A wrapper that removes instance_id in default crowd object factory."""
  del instance_id
  return object_factory(*args, **kwargs)


@gin.configurable
def random_object_factory(
    instance_id: int,
    object_factories: Iterable[
        Callable[..., autonomous_object.AutonomousObject]],
    *args, **kwargs) -> autonomous_object.AutonomousObject:
  """A wrapper that removes instance_id in default crowd object factory."""
  del instance_id
  object_factory = np.random.choice(object_factories)
  return object_factory(*args, **kwargs)


@gin.configurable
def sensor_factory(instance_id: int, sensor: Callable[...,
                                                      generic_sensor.Sensor],
                   *args, **kwargs) -> generic_sensor.Sensor:
  del instance_id
  return sensor(*args, **kwargs)


@gin.configurable
class CrowdBuilder(object):
  """A helper class to construct a crowd."""

  def __init__(
      self,
      num_instance: int,
      crowd_controller_factory: Callable[..., CrowdController],
      object_factory: Callable[..., autonomous_object.AutonomousObject],
      sensor_factories: Iterable[Callable[..., generic_sensor.Sensor]] = None):
    """Constructor.

    Args:
      num_instance: Number of autonomous objects in the crowd.
      crowd_controller_factory: A callable that returns a crowd controller
        object.
      object_factory: Callable that returns an autonomous object.
      sensor_factories: list of sensor callables.
    """
    self._objects = []
    crowd_id_prefix = "crowd"
    names = [crowd_id_prefix + "_%d" % i for i in range(num_instance)]

    self._controller = crowd_controller_factory(names=names)

    for i in range(num_instance):
      position_sensor = base_position_sensor.BasePositionSensor(
          name=names[i] + POSITION_SENSOR_POSTFIX)

      # Add additional per agent sensors (e.g. camera, occupancy, etc.).
      add_sensors = []
      if sensor_factories:
        for s in sensor_factories:
          add_sensors.append(
              sensor_factory(
                  instance_id=i, sensor=s, name=names[i] + "_" + s.__name__))

      an_object = object_factory(
          instance_id=i,
          sensors=(position_sensor,) + tuple(add_sensors),
          controller=self._controller.instance_controller(i))

      self._objects.append(an_object)

  @property
  def crowd_objects(self) -> List[autonomous_object.AutonomousObject]:
    """Returns list of AutonomousObjects in the crowd."""
    return self._objects

  @property
  def crowd_controller(self) -> CrowdController:
    """Returns the crowd controller."""
    return self._controller



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: minitaur_logging.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pybullet_envs.minitaur.envs import timestamp_pb2 as timestamp__pb2
from pybullet_envs.minitaur.envs import  vector_pb2 as vector__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='minitaur_logging.proto',
  package='robotics.reinforcement_learning.minitaur.envs',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16minitaur_logging.proto\x12-robotics.reinforcement_learning.minitaur.envs\x1a\x0ftimestamp.proto\x1a\x0cvector.proto\"k\n\x0fMinitaurEpisode\x12X\n\x0cstate_action\x18\x01 \x03(\x0b\x32\x42.robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction\"U\n\x12MinitaurMotorState\x12\r\n\x05\x61ngle\x18\x01 \x01(\x01\x12\x10\n\x08velocity\x18\x02 \x01(\x01\x12\x0e\n\x06torque\x18\x03 \x01(\x01\x12\x0e\n\x06\x61\x63tion\x18\x04 \x01(\x01\"\xd0\x02\n\x13MinitaurStateAction\x12\x12\n\ninfo_valid\x18\x06 \x01(\x08\x12*\n\x04time\x18\x01 \x01(\x0b\x32\x1c.robotics.messages.Timestamp\x12\x32\n\rbase_position\x18\x02 \x01(\x0b\x32\x1b.robotics.messages.Vector3d\x12\x35\n\x10\x62\x61se_orientation\x18\x03 \x01(\x0b\x32\x1b.robotics.messages.Vector3d\x12\x35\n\x10\x62\x61se_angular_vel\x18\x04 \x01(\x0b\x32\x1b.robotics.messages.Vector3d\x12W\n\x0cmotor_states\x18\x05 \x03(\x0b\x32\x41.robotics.reinforcement_learning.minitaur.envs.MinitaurMotorStateb\x06proto3'
  ,
  dependencies=[timestamp__pb2.DESCRIPTOR,vector__pb2.DESCRIPTOR,])




_MINITAUREPISODE = _descriptor.Descriptor(
  name='MinitaurEpisode',
  full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurEpisode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state_action', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurEpisode.state_action', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=211,
)


_MINITAURMOTORSTATE = _descriptor.Descriptor(
  name='MinitaurMotorState',
  full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='angle', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState.angle', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='velocity', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState.velocity', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='torque', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState.torque', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='action', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState.action', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=213,
  serialized_end=298,
)


_MINITAURSTATEACTION = _descriptor.Descriptor(
  name='MinitaurStateAction',
  full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='info_valid', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.info_valid', index=0,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.time', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_position', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.base_position', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_orientation', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.base_orientation', index=3,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_angular_vel', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.base_angular_vel', index=4,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='motor_states', full_name='robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction.motor_states', index=5,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=301,
  serialized_end=637,
)

_MINITAUREPISODE.fields_by_name['state_action'].message_type = _MINITAURSTATEACTION
_MINITAURSTATEACTION.fields_by_name['time'].message_type = timestamp__pb2._TIMESTAMP
_MINITAURSTATEACTION.fields_by_name['base_position'].message_type = vector__pb2._VECTOR3D
_MINITAURSTATEACTION.fields_by_name['base_orientation'].message_type = vector__pb2._VECTOR3D
_MINITAURSTATEACTION.fields_by_name['base_angular_vel'].message_type = vector__pb2._VECTOR3D
_MINITAURSTATEACTION.fields_by_name['motor_states'].message_type = _MINITAURMOTORSTATE
DESCRIPTOR.message_types_by_name['MinitaurEpisode'] = _MINITAUREPISODE
DESCRIPTOR.message_types_by_name['MinitaurMotorState'] = _MINITAURMOTORSTATE
DESCRIPTOR.message_types_by_name['MinitaurStateAction'] = _MINITAURSTATEACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MinitaurEpisode = _reflection.GeneratedProtocolMessageType('MinitaurEpisode', (_message.Message,), {
  'DESCRIPTOR' : _MINITAUREPISODE,
  '__module__' : 'minitaur_logging_pb2'
  # @@protoc_insertion_point(class_scope:robotics.reinforcement_learning.minitaur.envs.MinitaurEpisode)
  })
_sym_db.RegisterMessage(MinitaurEpisode)

MinitaurMotorState = _reflection.GeneratedProtocolMessageType('MinitaurMotorState', (_message.Message,), {
  'DESCRIPTOR' : _MINITAURMOTORSTATE,
  '__module__' : 'minitaur_logging_pb2'
  # @@protoc_insertion_point(class_scope:robotics.reinforcement_learning.minitaur.envs.MinitaurMotorState)
  })
_sym_db.RegisterMessage(MinitaurMotorState)

MinitaurStateAction = _reflection.GeneratedProtocolMessageType('MinitaurStateAction', (_message.Message,), {
  'DESCRIPTOR' : _MINITAURSTATEACTION,
  '__module__' : 'minitaur_logging_pb2'
  # @@protoc_insertion_point(class_scope:robotics.reinforcement_learning.minitaur.envs.MinitaurStateAction)
  })
_sym_db.RegisterMessage(MinitaurStateAction)


# @@protoc_insertion_point(module_scope)



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import pybullet_data
import pybullet as p
import time

p.connect(p.GUI_SERVER)
p.setAdditionalSearchPath(pybullet_data.getDataPath())

while (1):
  #this is a no-op command, to allow GUI updates on Mac OSX (main thread)
  p.setPhysicsEngineParameter()
  time.sleep(0.01)
